
/*
** Author : Steven Ramsay <smr@dcs.ed.ac.uk>, 15/02/97
** Files  : Coord.java, Coord.class
** Update : 11/06/97 
** Package: simdap
**
** Description:
** The coordinator communicates with other sites via the scheduler.
** All messages are filtered through the Site.
*/

package eduni.simdcp;

import eduni.simjava.*;
import eduni.simanim.*;

/* Methods defined in final class Coord
** PUBLIC
** Coord(String, int, int, int, String)
**        void    body() 
**
** PROTECTED
**
** PRIVATE
**        void    Init_Variables()
**        void    Create_Lists() 
**        boolean Handle_Start_Txn()
**        boolean Handle_End_Txn()
**        void    Set_Decision()
**        void    Begin_Txn_COORD(int)
**        int     Decide_Choice_COORD(States)
**        int     Handle_Choice_COORD(Choices)
**        int     Handle_Action_COORD(Actable)
**        int     Handle_Action_Msg_COORD(Msg)
**        int     Handle_Action_Ctrl_COORD(Ctrl)
**        int     Handle_Action_Log_COORD(Log)
**        Message Receive_Message()
**        void    Send_Message(Msg)
**        void    Send_Multicast_Msg(Message)
**        void    Schedule_Message(Message)
*/
final class Coord extends Site implements Constants {

  private Sim_port   port_site;                         // Port to site filter

  private Site_Lists Lists;

  public Coord(String name, int id, int x, int y, 
                                    String i_gif, String p_gif) {
      // Constructor for Coordinator at site
      super(name, id, COORD, x, y, i_gif, p_gif); 

      Debug("Initiating Coord at site "+Id());

      this.port_site = new Sim_port("port_Site"+id, p_gif,
                                    Option.Port_Side(COORD, id, SITE), 
                                    Option.Port_Pos(COORD, id));
      add_port(port_site);

      add_param(new Anim_param("State", Anim_param.VALUE, "", 
               Option.State_Pos_x(COORD, id), Option.State_Pos_y(COORD, id)));
      add_param(new Anim_param("Status", Anim_param.STATE, "sleep", 0, 0));

  }                                                     // constructor Site(2)
  
  protected void Init_Variables() {
    // Method to initialise the class variables because SimAnim does not 
    // destroy and recreate the objects.
      super.Init_Variables();

      this.Create_Lists();
      super.Set_Status(SLEEPING);
  }                                                  // void Init_Variables(0)

  public void body() {
      // Body for coordinator at site
      int     start_state_num = 1;
      boolean stop;

      this.Init_Variables();

      stop = this.Handle_Start_Txn();

      while (!stop) {
  	  if (Is_Coordinator()) {
              super.Show_Window();
              Info("-> Started Coordinator "+Id());
              super.Set_Status(RUN_OK);
              this.Begin_Txn_COORD(start_state_num);
          }
          else {
              super.Set_Status(SLEEPING);
              super.Hide_Window();
          }
          stop = this.Handle_End_Txn();
      }

      if (super.Status() == FAILED) {
          Info("<- Coordinator "+Id()+": "
               +Global.Name_CorA(super.Global_Decision())
               +" recovered after TXN completed");
      }
      Info("<- Terminated Coordinator "+Id()+": "
                                  +Global.Name_CorA(super.Global_Decision()));
      
  }                                                            // void body(0)

  private void Create_Lists() {
      // Method to create the lists for a coordinator
      this.Lists = new Site_Lists(COORD, Id(), this.Site_Window());
  }                                                    // void Create_Lists(0)

  private boolean Handle_Start_Txn() {
      // Method to handle the start of a txn - initiates Coord if required
      // Returns true if any error occured and txn not started
      boolean stop = true;
      Message msg, new_msg;

      msg = this.Receive_Message();

      if ((msg.Round() == -1) && (msg.Name().compareTo("PARTICIPATE") == 0)) {
          if (msg.Info() == YES) {
              super.Set_Role_Of_Site(COORD);
          }
          else {
              super.Hide_Window();
          }
          stop = false;
      }
      else {
          Error("First message does not start txn");
          stop = true;
      }
      super.Store_Current_Msg(null);

      return stop;
  }                                             // boolean Handle_Start_Txn(0)

  private boolean Handle_End_Txn() {
      // Method to handle the end of a txn - either terminate entity
      // Returns true if site terminated
      boolean stop = false;
      Message msg, new_msg;

      while (!stop) {
          msg = this.Receive_Message();

          if ((msg.Round() == -1) && (msg.Name().compareTo("END-TXN") == 0)) {
              // Transaction terminated
              if (Is_Coordinator()) {
	          if (super.Global_Decision() != msg.Info()) {
                      Error("Global decision has changed");
                  }
              }
              else {
                  super.Set_Global_Decision(msg.Info());
              } 
              stop = true;
              if (super.Global_Decision() != UNKNOWN) {
                  Info("<- TXN completed.");
              }
              Info("<- Global decision: "
                                  +Global.Name_CorA(super.Global_Decision()));
	  } 
          else {
              Info("End of TXN - message "+msg.Name()+" ignored");
              super.Store_Current_Msg(null);
          }
      }
      return stop;
  }                                               // boolean Handle_End_Txn(0)

  private void Set_Decision() {
      // Method to set the decision from the initial participant list.
      int id, decision, reply;
	  
      if (this.Lists.Num_Reply_To_Parts() 
                                    != this.Lists.Num_Replies_From_Parts()) {
          Error("Cannot decide - not all replies in");
      }
      else {
          decision = UNKNOWN;
          for (id=1; id <= Option.Num_Sites(); id++) {
	      if (this.Lists.Was_Part_Sent_Msg(id)) {
                  reply = this.Lists.Has_Part_Replied(id);
	          if (reply == UNKNOWN) {
                      Error("Part"+id+" not replied");
	          }
	          else if (reply == COMMIT) {
	              if (decision != ABORT) {
                          Debug("Setting decision COMMIT"); 
                          decision = COMMIT;
                      }
  	          }
	          else if (reply == ABORT) {
                      Debug("Setting decision ABORT"); 
                      decision = ABORT;
   	          }
                  Debug("reply "+reply+", decision "+decision); 
              }
          }
          if (decision == UNKNOWN) {
              decision = super.Global_Decision();
              Debug("Using previous global decision: "+decision);
          }
          super.Set_Global_Decision(decision);
      }
  }                                                    // void Set_Decision(0)


  private void Begin_Txn_COORD(int state_num) {
      // Send/wait for messages to/from participant as detailed in table,
      // given starting state.
      int     next_state_num;
      boolean stop;
      States  cur_state;

      stop = false;
      while ((state_num <= Table().Total_States(COORD)) && (!stop)) {
          super.Set_State_Num(state_num);

          cur_state = Table().State(COORD, state_num);
          next_state_num = this.Decide_Choice_COORD(cur_state);

          Debug("in state "+state_num); 
          Debug("next state "+next_state_num);
 
          if (next_state_num == 0) {
	      stop = true;                      // No state to move to so stop
          }
          else if (next_state_num == FAIL) {
              Debug("failed choices in state "+state_num);
              Error("<- TERMINATED BEFORE COMPLETION");
              stop = true;
          }
          else {
              state_num = next_state_num;
          }
      }
  }                                                 // void Begin_Txn_Coord(1)

  private int Decide_Choice_COORD(States state) {
      // Method to select a choice of actions given a state.
      // If there are no choices or only one then there is no decision to
      // be made, otherwise the choice will be made depending on the first 
      // action of each choice.  It may be simpler to try the first choice,
      // then if its first action fails the second choice could be tried.
      // If a first action succeeds then it is committed to that choice.
      // The next state will be returned (from the choice selected) or zero
      // to stop the txn.
      
      int next_state_num = 0;
      int total_choices;
      int choice_num;

      total_choices = state.Total_Choices();

      if (total_choices == 0) {                        // No choices available
          Error("no choices, stop");
          next_state_num = FAIL; 
      }
      else {                  // This will try each choice until it can do one
          next_state_num = RETRY;
          choice_num = 1;
	  while (next_state_num == RETRY) { 
	      if (choice_num > total_choices) {          // If no choices left
                  next_state_num = FAIL;
              }
              else {                                             // Try choice
                  Debug("trying choice "+choice_num);
                  super.Set_Choice_Num(choice_num);
                  next_state_num = 
                           this.Handle_Choice_COORD(state.Choice(choice_num));
              }
              choice_num++;
          }                       
      }
      return next_state_num;               // return next state number or fail
  }                                              // int Decide_Choice_Coord(1)

  private int Handle_Choice_COORD(Choices choice) {
      // Method to handle the choice of actions and return the next state. 
      int action_num;
      int success;
      int next_state_num;
      
      action_num = 1;
      success = OK;

      while ((action_num <= choice.Total_Actions()) && (success == OK)) {

          success = Handle_Action_COORD(choice.Action(action_num));

          if (success == OK) {
                               // Firt action okay so committed to this choice
              Debug("action "+action_num+" completed successfully");
              action_num++;
          }
          else if (action_num == 1) {
  	      // Could not handle first action so return to try another choice
	      // Leave stored message for use in trying next choice
              success = RETRY;
          }     
          else {                // Handled first action so committed to choice
              success = FAIL;
          }
      }
      if (success == OK) {
          next_state_num = choice.Next_State();
      }
      else { 
          next_state_num = success;
      }
      return next_state_num;                             // Return next state
  }                                              // int Handle_Choice_Coord(1)

  private int Handle_Action_COORD(Actable action) {
      // Method to handle the Coordinator's action depending on what it is,
      // i.e. message to be sent/received, control action, log r/w.
      int success;

      success = FAIL;

      Debug("handling action - "+action.Display());

      if (action instanceof Msg) {                        // Action is message
          success = this.Handle_Action_Msg_COORD((Msg) action);
      }

      else if (action instanceof Ctrl) {
          success = this.Handle_Action_Ctrl_COORD((Ctrl) action);
      }

      else if (action instanceof Log) {
          success = this.Handle_Action_Log_COORD((Log) action);
      }

      else {
          Error("unknown class");
      }
      Debug("returning success "+success); 
      return success;
  }                                              // int Handle_Action_COORD(1)

  private int Handle_Action_Msg_COORD(Msg msg_act) {
      // Method to handle a Coordinator's message action.
      // i.e. protocol 'real' message to be sent/received.
      int       success;
      Message   msg;

      success = FAIL;

      if (msg_act.Send()) {                                    // Send message
          this.Send_Message(msg_act);
          success = OK;
      }
      else if (msg_act.Receive()) {                         // Receive message
          Debug("expecting message "+msg_act.Name()+","+msg_act.Round());

	  msg = this.Receive_Message();

          if (msg.Src_Type() == PART) {
	      // Message is from participant
              if ((msg.Name().compareTo(msg_act.Name()) == 0) 
                                       && (msg_act.Round() == msg.Round())) {
      	          // Matching message pair - as expected
                  success = OK;
                  if (msg.Round() == super.Local_Msg_Round()) {
 	   	      // Replies must match the local msg round
		      // No other protocol 'REAL' message can match msg round
		      // Timeouts and control messages are handled differently
		      // so they may and will match local msg round.
		      // This ensures delayed messages not counted as replies 
                      Debug("Message included in replies");
                      this.Lists.Set_Part_Replied(msg.Src_Id(), msg.Info());
                  }
                  else {
		      // Delayed 'real' messages not with local round ignored
                      Debug("Message round differs - excluded as reply");
		  }
              } 
              else {
	          // Not a pair - rec message was not expected
                  Debug("expecting message "
                                        +msg_act.Name()+","+msg_act.Round());
                  Debug("received message "+msg.Name()+","+msg.Round());
              }
          }
          else if (msg.Src_Type() == SCHED) {
	      // Message is from Sched. If round 0 it will be to start txn.
              if ((msg.Round() == -5) && (msg.List())) {  
                  // Set Parts In Txn List and setup reply to list 
	          this.Lists.Copy_Parts_In_Txn_List(msg.Part_List());
                  this.Lists.Update_Reply_To_Parts_List(
                                              this.Lists.Parts_In_Txn_List());
                  success = OK;
              }
              else if (msg.Round() == -1) {
		  // Message was from Sched because simulation blocked
  	  	  // Do nothing
              }
              else {
		  // No other messages from Sched are handled here
	          Error("Unknown message from Scheduler"); 
              }
	  }
          else if (msg.Src_Type() == COORD) {  
              // Message is from another Coordinator
              Error("cannot handle Coord->Coord messages");
          }
          else {
              Error("Src_Type in Handle_Action_Msg_COORD(1)");
          }                  

          if (success == OK) { 
	      // Finished with message if it was expected
	      // Otherwise, must store for remaining choices
              super.Store_Current_Msg(null); 
          }
      }
      else {
          Error("can't handle message");
      }
      return success;
  }                                          // int Handle_Action_Msg_COORD(1)

  private int Handle_Action_Ctrl_COORD(Ctrl ctrl_act) {
      // Method to handle a Coordinator's Control action.
      // i.e. vote, timeout, commit or abort (end).
      int       success;
      int       num_replies_from, num_reply_to;
      Msg       msg_tmp;
      Message   msg;

      success = FAIL;

      if (ctrl_act.Name().compareTo("END") == 0) {
          num_replies_from = this.Lists.Num_Replies_From_Parts();
          num_reply_to     = this.Lists.Num_Reply_To_Parts();
          if (num_replies_from == num_reply_to) {
              Debug("Received all replies R"
                                         +num_replies_from+"=P"+num_reply_to);
              this.Set_Decision();
   	      if (super.Global_Decision() == ctrl_act.Info()) {
                  this.Lists.Update_Reply_To_Parts_List(
                                        this.Lists.Replies_From_Parts_List());
                  this.Lists.Reset_Replies_From_Parts_List();
                  success = OK;
              }
          }
          else {                   // At least one participant not replied
              Debug("Not received all replies R"
                                      +num_replies_from+" == P"+num_reply_to);
          }
      }

      else if (ctrl_act.Name().compareTo("TIMEOUT") == 0) {
          msg = Receive_Message();
          if (msg.Name().compareTo("TIMEOUT") == 0) {
	      if (ctrl_act.Msg().Round() >= msg.Round()) {
                  success = OK;
                  Info("TIMEOUT in state "+super.State_Num()
                       +":"+super.State_Name());
                  Global.Inc_Timeouts_COORD();
		  // Assume all timeouts current or previous are valid
                  this.Lists.Set_Part_Replied(msg.Src_Id(), FAILED);
	      }
              else {
                  Debug("ignoring timeout, ctrl msg rnd "
                        +ctrl_act.Msg().Round()+", message rnd "
                        +msg.Round()+" - not current");
		  // Lose all non-current timeout messages
                  super.Store_Current_Msg(null); 
	      }
	  }
	  else {
              Debug("Message received not TIMEOUT");
	  }
          if (success == OK) { 
              super.Store_Current_Msg(null); 
          }
      }
 
      else if (   (ctrl_act.Name().compareTo("COMMIT") == 0) 
               || (ctrl_act.Name().compareTo("ABORT")  == 0)) {
  	       
          if (ctrl_act.Is_Msg()) {
              msg_tmp = ctrl_act.Msg();
              if (msg_tmp.Send()) {
                  msg = new Message(ctrl_act.Name(), SCHED, SCHED, null, 
                              COORD, Id(), 0, ctrl_act.Info(), null, true);
                  this.Schedule_Message(msg);
                  success = OK;
              }
              else {
                  Error("can't handle receive END control message");
              }
          }
          else {
              Error("invalid Ctrl action for commit or abort - no message");
          }
      }
                 
      else {
          Error("unknown ctrl message");
      }

      return success;
  }                                         // int Handle_Action_Ctrl_COORD(1)

  private int Handle_Action_Log_COORD(Log log_act) {
      // Method to handle a Coordinator's log action.
      // i.e. log read or write.
      int       success, info;
      String    name;
      Log       log_entry;

      success = FAIL;

      if (log_act.RW() == W) {
          // Write log information to DT log
          this.Lists.Write_DT_Log(log_act);
          Info("DT Log write: "+log_act.Name()
               +" ("+Global.Name_CorA(log_act.Info())+")");
          success = OK;
      }
      else if (log_act.RW() == R) {
	  // Read last DT log entry and match with expected log action
          log_entry = this.Lists.Read_DT_Log();
          name = log_entry.Name();
          info = log_entry.Info();
          if ((log_act.Name().compareTo(name) == 0) 
                                               && (log_act.Info() == info)) {
              Info("DT Log read: "+name+" ("+Global.Name_CorA(info)+")");
              success = OK;
          }
          else {
              Debug("log entry ("+name+", "+info+") expected, read ("
                    +log_act.RW()+", "+log_act.Info()+")");
          }
      }
      else {
          Error("Unknown log access");
      }
      return success;
  }                                          // int Handle_Action_Log_COORD(1)

  private Message Receive_Message() { 
      // Method to wait for a message (event) or return the first queued
      // in buffer.  The buffered message will be returned first.
      Message   msg = null;
      Sim_event ev = new Sim_event();

      msg = super.Stored_Message();

      if (msg != null) {            // Only wait for message if not one stored
          Debug("using stored message");
      }
      else {
          Debug("No message currently stored"); 
          Debug("number of events queued: "+sim_waiting());
          Debug("waiting for event to arrive");

          sim_get_next(ev);

          msg = (Message) ev.get_data();

          Trace_Msg_Received(msg);

          super.Store_Current_Msg(msg);         // Store current message

          if (msg.Control()) {
   	      if (msg.Round() == -5) {
                  Info("Received START");
	      }
              Global.Inc_Ctrl_Msgs_Rec();
              Debug("control msgs received "+Global.Ctrl_Msgs_Rec());
          }
          else {
              Info("Received "+msg.Name());
              Global.Inc_Real_Msgs_Rec();
              Debug("real msgs received "+Global.Real_Msgs_Rec());
          }
      }
      if (msg.Round() == (super.Local_Msg_Round() + 1)) {
          super.Inc_Local_Msg_Round();
      }
      return msg;
  }                                              // Message Receive_Message(0)

  private void Send_Message(Msg msg_act) {
      // Method to send a message given the source type and action
      // describing message.
      // Assumes coord messages are for part only and vice versa.
      int[]   part_list = new int[MAX_SITES+1];
      int[]   dest_list = new int[MAX_SITES+1];
      Message msg;

      Debug("about to send message");
   
      if (msg_act.Round() == (Global.Message_Round() + 1)) {
          Global.Inc_Msg_Round();
      }

      if (msg_act.Round() == (super.Local_Msg_Round() + 1)) {
          super.Inc_Local_Msg_Round();
      }

      if (msg_act.Multicast()) {
          // Send message to all participants in dest_list[]
          dest_list = this.Lists.Reply_To_Parts_List();
          part_list = null;         
          if (msg_act.Round() == 1) {
              part_list = this.Lists.Parts_In_Txn_List();
          }
          // Send list of participants to each part.
          msg = new Message(msg_act.Name(), PART, UNKNOWN, dest_list, 
                            COORD, Id(), msg_act.Round(), 
                            msg_act.Info(), part_list);

          this.Send_Multicast_Msg(msg, COMMIT);
      }
      else {
          dest_list = this.Lists.Replies_From_Parts_List();
          Debug("Must send "+msg_act.Name()+" to the following parts:");
          this.Lists.Display_Parts_In_Reply_List();
          msg = new Message(msg_act.Name(), PART, UNKNOWN, dest_list,
                            COORD, Id(), msg_act.Round(),
                            msg_act.Info(), null);

          this.Send_Multicast_Msg(msg, FAILED);
      }
  }                                                    // void Send_Message(2)

  private void Send_Multicast_Msg(Message snd_msg, int info) {
      // Method to handle a multicast messages
      int     id, msg_count;
      int[]   dest_list = new int[MAX_SITES+1];
      Message msg, last_msg;

      msg = snd_msg.Clone();
      msg_count = 0;

      Debug("about to send multicast message");

      dest_list = msg.Dest_List();

      for (id=1; id <= Option.Num_Sites(); id++) {
          msg = snd_msg.Clone();
          if (Global.In_List(dest_list, id, info)) {
              msg.Set_Dest_Id(id);
              Debug("Sending to part "+id+", info "+Global.Name_CorA(info));
              if (info == FAILED) {
  	    	  // The participant may recover when gets message
		  this.Lists.Set_Part_Replied(id, UNKNOWN);
	      }
              msg_count++;
              this.Schedule_Message(msg);
          }
      }
      Debug("Sent "+msg_count+" messages in round "+msg.Round());
      /* Removed as not required - smr
      ** last_msg = new Message("LAST_MSG", SCHED, SCHED, msg.Dest_List(), 
      **                        COORD, Id(), msg.Round(),
      **                        msg_count, null);  
      ** this.Schedule_Message(last_msg);
      */
  }                                              // void Send_Multicast_Msg(2)

  private void Schedule_Message(Message snd_msg) {
      // Method to actually send a message to intended recipient. 
      Message msg;

      msg = snd_msg.Clone();
      sim_schedule(port_site, LINK_DELAY, SITE, msg);

      Trace_Msg_Sent(msg, port_site.get_pname());

      if (msg.Control()) {
          Global.Inc_Ctrl_Msgs_Sent();
          Debug("control msgs sent "+Global.Ctrl_Msgs_Sent());
      }
      else {
          Info("Sent "+msg.Name()+" (state "+super.State_Num()
               +":"+super.State_Name()+")");
          Global.Inc_Real_Msgs_Sent();
          Debug("real msgs sent "+Global.Real_Msgs_Sent());
      }
  }                                                // void Schedule_Message(1)

}                                                               // class Coord

/*---------------------------- End of Coord.java ----------------------------*/

