
/*
** Author : Steven Ramsay <smr@dcs.ed.ac.uk>, 16/02/97
** Files  : Filter.java, Filter.class
** Update : 11/06/97 
** Package: simdcp
**
** Description:
** The site simply forwards messages from Scheduler to Coord/Part
** and vice versa.
*/

package eduni.simdcp;

import eduni.simjava.*;
import eduni.simanim.*;

/* Methods defined in final class Filter
** PUBLIC
** Filter(String, int, int, int, String)
**        void    body() 
**
** PROTECTED
**
** PRIVATE
**        void    Init_Variables()
**        boolean Handle_Start_Txn()
**        boolean Handle_Filter_Msg(Message)
**        Message Receive_Message()
**        void    Forward_Message(Message)
**        void    Send_Message(Message)
*/
final class Filter extends Site implements Constants {
  private Sim_port port_sched;                            // Port to Scheduler
  private Sim_port port_coord;                          // Port to Coordinator
  private Sim_port port_part;                           // Port to Participant

  public Filter(String name, int id, int x, int y, 
                                     String i_gif, String p_gif) {
      // Constructor for site Filter
      super(name, id, SITE, x, y, i_gif, p_gif);

      Debug("Initiating Filter at site "+Id());

      this.port_sched = new Sim_port("port_Sched", p_gif,
                                     Option.Port_Side(SITE, id, SCHED), 
                                     Option.Port_Pos(SITE, id));
      this.port_coord = new Sim_port("port_Coord"+id, p_gif,
				     Option.Port_Side(SITE, id, COORD), 
				     Option.Port_Pos(SITE, id));
      this.port_part  = new Sim_port("port_Part"+id, p_gif,
                                     Option.Port_Side(SITE, id, PART), 
                                     Option.Port_Pos(SITE, id));
      add_port(port_sched);
      add_port(port_coord);
      add_port(port_part);

      add_param(new Anim_param(" "+Id(), Anim_param.NAME, "id", 
                                  Option.Id_Pos_x(id), Option.Id_Pos_y(id)));
  }                                                   // constructor Filter(2)
  
  protected void Init_Variables() {
    // Method to initialise the class variables because SimAnim does not 
    // destroy and recreate the objects.
      super.Init_Variables();
  }                                                  // void Init_Variables(0)

  public void body() {
      // Body for filter at site
      int     count = 0;
      boolean stop;
      Message msg;

      this.Init_Variables();

      stop = this.Handle_Start_Txn();
 
      super.Show_Window();
      Info("-> Started site Filter "+Id());
      while (!stop) {
  	   msg = this.Receive_Message();
           if ((msg.Dest_Type() == SITE) && (msg.Dest_Id() == Id())) {
	       // Handle messages for this Site 
               stop = this.Handle_Filter_Msg(msg);
           }
           else {
               this.Forward_Message(msg);
           }
      }
      Info("<- Terminated site Filter "+Id()+": "
                                  +Global.Name_CorA(super.Global_Decision()));
  }                                                            // void body(0)

  private boolean Handle_Start_Txn() {
      // Method to handle the start of a txn - forwards Participate message
      // to Participant and Coordinator.
      // Returns true if any error occured and txn not started
      int role     = UNKNOWN;
      boolean stop = true;
      Message msg, new_msg;

      msg = this.Receive_Message();

      if ((msg.Round() == -1) && (msg.Name().compareTo("PARTICIPATE") == 0)) {
          super.Set_Role_Of_Site(msg.Info());
          new_msg = msg.Clone();

          if (Is_Coordinator()) {                    // Site has a coordinator
              new_msg.Set_Info(YES);
          }
          else {
  	      new_msg.Set_Info(NO);
          }
          new_msg.Set_Dest_Type(COORD);
          this.Send_Message(new_msg);

          new_msg = msg.Clone();

          if (Is_Participant()) {                    // Site has a participant
 	      if (Global.In_List(msg.Part_List(), Id(), COMMIT)) {
                  new_msg.Set_Info(COMMIT);
                  super.Set_Local_Decision(COMMIT);
              }
 	      else if (Global.In_List(msg.Part_List(), Id(), ABORT)) {
                  new_msg.Set_Info(ABORT);
                  super.Set_Local_Decision(ABORT);
              }
          }
          else {
  	      new_msg.Set_Info(NONE);
          }
          new_msg.Set_Dest_Type(PART);
          this.Send_Message(new_msg);
          stop = false;
      }
      else {
          Error("First message does not start txn");
          stop = true;
      }
      return stop;
  }                                              // boolean Handle_Start_Txn()

  private boolean Handle_Filter_Msg(Message msg) {
      // Method to handle a message for the Site (Filter)
      // Returns true if it ends txn
      boolean stop = false;
      Message new_msg1, new_msg2;

      if ((msg.Round() == -1) && (msg.Name().compareTo("END-TXN") == 0)) {
          // Transaction terminated
          super.Set_Global_Decision(msg.Info());
          new_msg1 = msg.Clone();
          new_msg1.Set_Dest_Type(COORD);
          this.Send_Message(new_msg1); 

          new_msg2 = msg.Clone();
          new_msg2.Set_Dest_Type(PART);
          this.Send_Message(new_msg2); 
          stop = true;
      }
      return stop;
  }                                            // boolean Handle_Filter_Msg(1)

  private Message Receive_Message() { 
      // Method to return the first message in deferred queue or
      // wait for event to arrive.
      int       num_waiting;
      Message   msg = null;
      Sim_event ev = new Sim_event();

      num_waiting = sim_waiting();
      if (num_waiting > 0) {
          Route("num of events queued : "+num_waiting);
          Route("num waiting for Coord: "+sim_waiting(new Sim_type_p(COORD)));
          Route("num waiting for Part : "+sim_waiting(new Sim_type_p(PART)));
          Route("num waiting for Sched: "+sim_waiting(new Sim_type_p(SCHED)));
      }
      else {
          Route("waiting for event to arrive");
      }
      sim_get_next(ev);

      msg = (Message) ev.get_data();

      Trace_Msg_Received(msg);

      if (msg.Round() == (super.Local_Msg_Round() + 1)) {
          super.Inc_Local_Msg_Round();
      }

      if (msg.Control()) {
	  // Received control message
          Global.Inc_Ctrl_Msgs_Rec();
          Route("control msgs received "+Global.Ctrl_Msgs_Rec());
      }
      return msg.Clone();
  }                                              // Message Receive_Message(0)

  private void Forward_Message(Message snd_msg) {
      // Method to forward a given message to the correct dest
      Message msg;

      msg = snd_msg.Clone();

      Send_Message(msg);
  }                                                 // void Forward_Message(1)

  private void Send_Message(Message snd_msg) {
      // Method to actually send a message to intended recipient. 
      int     dest_type, dest_id;
      Message msg;

      msg = snd_msg.Clone();
      dest_type = msg.Dest_Type();
      dest_id   = msg.Dest_Id();

      if (dest_id == Id()) {
	  // For this site 
	  if (dest_type == msg.Src_Type()) {
              Error("sender is also intended recipient at this site");
          }
  	  else if (dest_type == COORD) { 
              sim_schedule(port_coord, LINK_DELAY, dest_type, msg);
              Trace_Msg_Sent(msg, port_coord.get_pname());
          }
          else if (dest_type == PART) {
              sim_schedule(port_part, LINK_DELAY, dest_type, msg);
              Trace_Msg_Sent(msg, port_part.get_pname());
          }
          else {
              Error("cannot forward to type "+dest_type+" at this site");
          }
      }
      else {
          sim_schedule(port_sched, LINK_DELAY, SCHED, msg);
          Trace_Msg_Sent(msg, port_sched.get_pname());
      }

      if (msg.Control()) {
	  // Sent a control message
          Global.Inc_Ctrl_Msgs_Sent();
          Route("control msgs sent "+Global.Ctrl_Msgs_Sent());
      }
  }                                                    // void Send_Message(1)

}                                                              // class Filter

/*---------------------------- End of Filter.java ---------------------------*/

