
/*
** Author : Steven Ramsay <smr@dcs.ed.ac.uk>, 15/03/97
** Files  : Message.java, Message.class
** Package: simdcp
** Update : 11/06/97 
**
** Description:
** This is the structure of messages which are "piggybacked" on
** events between entities (Coord, Part, Filter and Sched)
*/

package eduni.simdcp;

/* Methods defined in final class Message
** PUBLIC
** Message(String, int, int, int[], int, int, int, int, int[])
** Message(String, int, int, int[], int, int, int, int, int[], boolean)
**
** PROTECTED
**        String  Name()
**        int     Src_Type()
**        int     Src_Id()
**        void    Set_Dest_Type(int)
**        int     Dest_Type()
**        void    Set_Dest_Id(int)
**        int     Dest_Id()
**        int[]   Dest_List()
**        int     Round()
**        void    Set_Info(int)
**        int     Info()
**        boolean Control()
**        boolean List()
**        int[]   Part_List()
**        Message Clone()
**        void    Display()
**        String  Sent_Data(int) 
**        String  Rec_Data(int) 
**
** PRIVATE
** static void    Error(String)
*/
final class Message extends Object implements Constants {
  // Control messages are sent between Scheduler and Coordinator/Participants.
  //
  // For source and destination, Scheduler is -1,
  // Coordinator and participants are given by their index in port_px[].
  private String  name;                                     // Name of message
  private int     src_type;             // Src type of message i.e. Coord/Part
  private int     dest_type;                // Dest type i.e. Coord/Part/Sched
  private int     src_id;                              // Source id of message
  private int     dest_id;                        // Destination id of message
  private int[]   dest_list = new int[MAX_SITES+1];     
                                                           // Destination list
  private int     msg_round;                        // Number of message round 
  private int     info;                 // Extra information e.g. Commit/Abort
  private int[]   part_list = new int[MAX_SITES+1];
                                              // List of participants involved 
  private boolean list;                      // Was a list of parts in message
  private boolean control;             // True if message is a control message

  public Message(String name, int dest_type, int dest_id, int[] dest_list, 
                 int src_type, int src_id, int msg_round, int info,
                 int[] p_list) {
      // Constructor for Message.  What the type of message is, its name
      // and source and destination of the message is set.
      this.name      = name;
      this.dest_type = dest_type;
      this.dest_id   = dest_id;
      this.src_type  = src_type;
      this.src_id    = src_id;
      this.msg_round = msg_round;
      this.info      = info;
      this.control   = false;

      if (dest_id == 0) {
          if (dest_list == null) {
              Error("Dest Id 0, yet no reply list given");
          }
          else {
              System.arraycopy(dest_list, 0, 
                               this.dest_list, 0, Option.Num_Sites()+1);
          }
      }
      if (p_list == null) {
          this.list = false;
      }
      else {
          System.arraycopy(p_list, 0, 
                           this.part_list, 0, Option.Num_Sites()+1);
          this.list = true;
      }
  }                                                  // constructor Message(9)

  public Message(String name, int dest_type, int dest_id, int[] dest_list, 
                 int src_type, int src_id, int msg_round, int info,
                 int[] p_list, boolean control) {
      // Constructor for Message. As for other constructor but adds 
      // a control field to distinguish control messages.
      this(name, dest_type, dest_id, dest_list, src_type, src_id, 
           msg_round, info, p_list);

      this.control = control;
  }                                                 // constructor Message(10)


  protected String Name() {
      // Public Method to return the name of a message.
      return this.name;
  }                                                          // String Name(0)

  protected int Src_Type() {
      // Public Method to return the src type of a message.
      return this.src_type;
  }                                                         // int Src_Type(0)

  protected int Src_Id() {
      // Public Method to return the source of a message.
      return this.src_id;
  }                                                           // int Src_Id(0)

  protected void Set_Dest_Type(int type) {
      // Public method to set the destination type 
      this.dest_type = type;
  }                                                   // void Set_Dest_Type(1)

  protected int Dest_Type() {
      // Public Method to return the dest type of a message.
      return this.dest_type;
  }                                                        // int Dest_Type(0)

  protected void Set_Dest_Id(int id) {
      // Public method to set the destination id 
      this.dest_id = id;
  }                                                     // void Set_Dest_Id(1)

  protected int Dest_Id() {
      // Public method to return the destination of a message.
      return this.dest_id;
  }                                                          // int Dest_Id(0)

  protected int[] Dest_List() {
      // Method to return the current reply list.
      return this.dest_list;
  }                                                      // int[] Dest_List(0)

  protected int Round() {
      // Public method to return the round a message belongs to.
      return this.msg_round;
  }                                                            // int Round(0)
  
  protected void Set_Info(int info) {
      // Public method to set the info
      this.info = info;
  }                                                        // void Set_Info(1)

  protected int Info() {
      // Public method to return the extra info passed with message.
      return this.info;
  }                                                             // int Info(0)

  protected boolean Control() {
      // Public method to return true if message is a control message
      return this.control;
  }                                                      // boolean Control(0)

  protected boolean List() {
      // Public method to indicate if a list was included in message.
      return this.list;
  }                                                         // boolesn List(0)

  protected int[] Part_List() {
      // Public Method to return the list of participants.
      int[] p_list = new int[MAX_SITES+1];

      System.arraycopy(this.part_list, 0, p_list, 0, Option.Num_Sites()+1);
      
      return p_list;
  }                                                      // int[] Part_List(0)


  protected Message Clone() {
      // Method to copy the message object rather than pass it by value.
      // .clone() doesn't seem to work
      Message msg;
      msg = new Message(this.Name(), 
                        this.Dest_Type(), this.Dest_Id(), this.Dest_List(), 
                        this.Src_Type(), this.Src_Id(),
                        this.Round(), this.Info(), 
                        this.Part_List(), this.control);
      return msg;
  }                                                        // Message Clone(0)

  protected void Display() {
      // Method to display the message  
      Scheduler.Debug_Pub("Message : "+this.Name()+","+this.Dest_Type()+","
                     +this.Dest_Id()+","+this.Src_Type()+","+this.Src_Id()
                     +","+this.Round()+","+this.Info());
  }                                                         // void Display(0)

  protected String Sent_Data(int type) {
      // Method to return the message as a string in Sent form
      int src_id    = this.Src_Id();
      int dest_id   = this.Dest_Id();
      int src_type  = this.Src_Type();
      int dest_type = this.Dest_Type();
      String note, dest_name, src_name;
      note = null;

      if (src_type == SCHED) {
          src_name = new String(Global.Short_Type(SCHED));
      }
      else {
          src_name = new String(Global.Short_Type(src_type)+src_id);
      }
      if (dest_type == SCHED) {
          dest_name = new String(Global.Short_Type(SCHED));
      }
      else {
	  if (dest_id != 0) {
              dest_name = new String(Global.Short_Type(dest_type)+dest_id);
          }
          else {
              dest_name = new String("multi "+Global.Short_Type(dest_type));
          }
      }

      if (src_type == type) {
          note = new String(this.Name()+" to "+dest_name+", from "+src_name
                            +", R:"+this.Round()+", I:"
                            +Global.Name_CorA(this.Info()));
      }
      else {
          note = new String(this.Name()+" to "+dest_name+", from "+src_name);
      }
      return note;
  }                                                     // String Sent_Data(1)

  protected String Rec_Data(int type) {
      // Method to return the message as a string in Rec form 
      int src_id    = this.Src_Id();
      int dest_id   = this.Dest_Id();
      int src_type  = this.Src_Type();
      int dest_type = this.Dest_Type();
      String note, dest_name, src_name;
      note = null;

      if (src_type == SCHED) {
          src_name = new String(Global.Short_Type(SCHED));
      }
      else {
          src_name = new String(Global.Short_Type(src_type)+src_id);
      }
      if (dest_type == SCHED) {
          dest_name = new String(Global.Short_Type(SCHED));
      }
      else {
	  if (dest_id != 0) {
              dest_name = new String(Global.Short_Type(dest_type)+dest_id);
          }
          else {
              dest_name = new String("multi "+Global.Short_Type(dest_type));
          }
      }
      if (dest_type == type) {
          note = new String(this.Name()+" from "+src_name+", for "+dest_name
                            +", R:"+this.Round()
                            +", I:"+Global.Name_CorA(this.Info()));
      }
      else {
          note = new String(this.Name()+" from "+src_name+", for "+dest_name);
      }
      return note;
  }                                                      // String Rec_Data(1)

  private static void Error(String note) {
      // Method to output error message if required
      Scheduler.Error_Pub(note+" (class Message)");
  }                                                           // void Error(1)

}                                                             // class Message

/*------------------------------ End of Message.java ------------------------*/

