
/*
** Author : Steven Ramsay <smr@dcs.ed.ac.uk>, 15/03/97
** Files  : Msg.java, Msg.class
** Package: simdcp
** Update : 11/06/97 
**
** Description:
** Structure of message actions in protocol table.
*/

package eduni.simdcp;

/* Methods defined in final class Msg
** PUBLIC
** Msg(String, int, int, int, int)
**        String  Name()
**        String  Display()
**        int     Info()
**
** PROTECTED
**        int     Round()
**        boolean Send()
**        boolean Receive()
**        boolean Unicast()
**        boolean Multicast()
**
** PRIVATE
**        void    Set_SR(int)
**        void    Set_UM(int)
**        char    Send_Rec()
**        char    Uni_Multi()
** static void    Error(String)
*/
final class Msg implements Actable, Constants {
  // This method implements the Actable interface and so it can be thought of
  // as a reference type.  This will allow the class Msg to be used 
  // interchangeably with other classes implementing this interface e.g. log 
  // writes.
  // A message will either be sent or received and can be unicast to/from one
  // site, multicast to/from many sites or broadcast to/from many sites.  The
  // broadcast option is not yet implemented.
  // Multicast means send to all participants who are in the list.
  private String msg_name;                          // The name of the message
  private int    info;                  // Extra information e.g. Commit/Abort
  private int    send_rec;                    // Is message to be rec or sent?
  private int    uni_mult;           // Are there multiple copies to send/rec?
  private int    round;                            // Round message belongs to
 
  public Msg(String name, int msg_round, int msg_sr, int msg_um, int info) {
      // Constructor for class
      this.msg_name = name;
      this.round    = msg_round;
      this.info     = info;
      this.Set_SR(msg_sr);
      this.Set_UM(msg_um);
  }                                                      // constructor Msg(5)

  public String Name() {
      return this.msg_name;
  }                                                          // String Name(0)

  public String Display() {
      // Method used in displaying the table
      String note = null;

      if (Option.Show_Debug()) {
          note = new String("M("+this.Send_Rec()+this.Round()+","
                            +this.Uni_Multi()+","+this.Info()+"):"
                            +this.Name());
      }
      else {
          note = new String("M:"+this.Name());
      }
      return note;
  }                                                       // String Display(0)

  public int Info() {
      // Public method to return the extra info in msg.
      return this.info;
  }                                                             // int Info(0)

  private void Set_SR(int msg_sr) {
      // Method to set the direction of message i.e. send or rec
      if (msg_sr == SND) {
          this.send_rec = SND;
      }
      else if (msg_sr == REC) {
          this.send_rec = REC;
      }
      else {
          Error("Message must be sent or received");
      } 
  }                                                          // void Set_SR(1)

  private void Set_UM(int msg_um) {
      // Method to set the number of messages sent/rec to/from the sites
      if (msg_um == UNI) {
          this.uni_mult = UNI;
      }
      else if (msg_um == MLT) {
          this.uni_mult = MLT;
      }
      else {
          Error("Message must be unicast or multicast.");
      } 
  }                                                          // void Set_UM(1)

  private char Send_Rec() {
      // Method to return a letter to indicate whether message is to be 
      // sent (S) or received (R) or neither (E)
      char SR = 'E';

      if (this.Send()) { 
          SR = 'S';
      }
      else if (this.Receive()) {
          SR = 'R';
      }
      return SR;
  }                                                        // char Send_Rec(0) 

  private char Uni_Multi() {
      // Method to return a letter to indicate whether message is to be 
      // unicast (U) or multicast (M) or neither (E)
      char UM = 'E';

      if (this.Unicast()) {
          UM = 'U';
      }
      else if (this.Multicast()) {
          UM = 'M';
      }
      return UM;
  }                                                       // char Uni_Multi(0)

  protected int Round() {
      // Method to return the round for a message
      return this.round;
  }                                                            // int Round(0)

  protected boolean Send() {
      // Method to determine if message is to be sent
      boolean bool = false;

      if (this.send_rec == SND) {
          bool = true;
      }
      return bool;     
  }                                                         // boolean Send(0)

  protected boolean Receive() {
      // Method to determine if message is to be received
      boolean bool = false;

      if (this.send_rec == REC) {
          bool = true;
      }
      return bool;     
  }                                                      // boolean Receive(0)

  protected boolean Unicast() {
      // Method to determine if only one message 
      boolean bool = false;

      if (this.uni_mult == UNI) {
          bool = true;
      }
      return bool;     
  }                                                      // boolean Unicast(0)

  protected boolean Multicast() {
      // Method to determine if several messages
      boolean bool = false;

      if (this.uni_mult == MLT) {
          bool = true;
      }
      return bool;     
  }                                                    // boolean Multicast(0)

  private static void Error(String note) {
      // Method to output error message if required
      Scheduler.Error_Pub(note+" (class Msg)");
  }                                                           // void Error(1)
}                                                                 // class Msg

/*------------------------------- End of Msg.java ---------------------------*/

