
/*
** Author : Steven Ramsay <smr@dcs.ed.ac.uk>, 15/01/97
** Files  : Scheduler.java, Scheduler.class
** Update : 11/06/97 
** Package: simdcp
**
** Description:
** The scheduler is used to control the message passing between the sites.
*/

package eduni.simdcp;

import eduni.simjava.*;
import eduni.simanim.*;

/* Methods defined in final class Scheduler 
** PUBLIC
** Scheduler(String, int, int, String)
**        void    body()
**
** PROTECTED
**        void    finalize() 
** static void    Debug_Pub(String)
** static void    Error_Pub(String)
** static void    Info_Pub(String)
**
** PRIVATE
** static void    Init_Variables()
** static void    Send_Participate_Msgs()
** static void    Send_FailSafe_Exit()
** static boolean Start_Txn()
** static boolean End_Txn()
** static boolean Handle_Sched_Msg(Message)
** static int     Num_Coords()
** static int     Num_Parts()
** static int     Num_CorA_Parts()
** static int     Num_Commited()
** static int     Num_Aborted()
** static void    Set_Port(int, String)
** static Sim_port Get_Port(int)
** static int     Get_Coord_Id()
** static void    Set_Part_In_Txn(int, int)
** static boolean Part_In_Txn(int)
** static int     Part_CorA(int)
** static void    Set_Part_CorA(int, int)
** static void    Set_x_Part_CorA(int, int)
** static void    Set_Coord_In_Txn(int, int)
** static boolean Coord_In_Txn(int)
** static void    Display_Coord_List()
** static void    Display_Part_List()
** static int[]   Part_List()
** static int[]   Part_CorAList()
** static void    Inc_Local_Msg_Round()
** static int     Local_Msg_Round()
** static void    Set_Global_Decision(int)
** static int     Global_Decision()
** static Table   Table()
** static void    Display_Statistics()
**        Message Receive_Message()
**        void    Forward_Message(Message)
**        void    Send_Message(Message)
**        void    Trace_Sent(Message)
**        void    Trace_Received(Message)
** static void    Setup_Window()
** static void    Reset_Window()
** static void    Show_Window()
** static void    Hide_Window()
** static void    Close_Window()
** static void    Debug(String)
** static void    Debug_GUI(String)
** static void    Error(String)
** static void    Error_GUI(String)
** static void    Info(String)
** static void    Info_GUI(String)
** static void    Route(String)
**        void    Trace_S(String, String)
**        void    Trace_P(String)
**        void    Trace(String)
*/
final class Scheduler extends Sim_entity implements Constants {

  // The participants ID is used when indexing into the array port_px[]
  // as each participant has a unique ID.  Indexing into the array
  // part_list[] with the same ID will indicate whether the participant is
  // to be sent a message, i.e. included in the list.  A similar list will be 
  // maintained for the coordinators, but it will return their IDs.  
  //
  // Note: It is possible for a part. and coord. to share the same site and
  // so have the same ID and port. 
  //
  // No participant/coordinator has ID = 0
  //
  // The participant with ID = COORD_ID will be assumed to be the coordinator 
  // and should be excluded from the participant list.
  //
  // port_site[Site ID] = port
  // part_list[Part. ID] = {no=exclusion, yes=inclusion}
  // part_CorA_list[Part. ID] = {COMMIT, ABORT, UNKNOWN}
  // coord_list[Coord ID] = {no=not active, yes=active} 

  // Set up an array of ports, one for each participant.
  private static Sim_port[] port_site;
  private static Sim_port   port_self;

  private static int[] part_list;
  private static int[] part_CorA_list;
  private static int[] coord_list;
  private static int[] timeout_list;

  private static int final_outcome;                // Did txn Commit or Abort?
  private static int local_msg_round;                   // Local message round

  private static W_Box  sched_window;

  private static int    x, y;
  private static String i_gif, p_gif;

  public Scheduler(String name, int x, int y, String i_gif, String p_gif) {
    // Method for creating an object of class Scheduler.
    super(name, i_gif, x, y); 
    this.x = x; this.y = y;
    this.i_gif = i_gif; this.p_gif = p_gif;
    int id;

    Debug("Initiating Scheduler");

    port_site = new Sim_port[Option.Num_Sites()+1];

    for (id=1; id <= Option.Num_Sites(); id++) {
        // For each participant id, create a port in port_px[id].
        Set_Port(id, "port_Site"+id, p_gif);    // Create port with given name
        add_port(Get_Port(id));                          // Add port to system
    }    

    this.port_self = new Sim_port("port_Self", p_gif, Anim_port.TOP, 28);
    add_port(port_self);

    add_param(new Anim_param(" Adversary", Anim_param.NAME, 
                             "A", SCHED_w, SCHED_h));
    add_param(new Anim_param("Global Decision", Anim_param.NAME_VALUE, 
                             "UNKNOWN", 0, Option.Off_y(SCHED_h+20)));
  }                                                // constructor Scheduler(1)
  
  protected void finalize() throws Throwable {
      // Finalize method to clean up windows opened etc when site terminates
      Close_Window();
      super.finalize();
  }                                                        // void finalize(0)

  private static void Init_Variables() {
    // Method to initialise the class variables because SimAnim does not 
    // destroy and recreate the objects.
    int id;

    part_list      = new int[Option.Num_Sites()+1];
    part_CorA_list = new int[Option.Num_Sites()+1];
    coord_list     = new int[Option.Num_Sites()+1];
    timeout_list   = new int[Option.Num_Sites()+1];

    final_outcome   = UNKNOWN;
    local_msg_round = 0;

    for (id=1; id <= Option.Num_Sites(); id++) {
        // For each participant id, add part to list and set to Commit
        Set_Part_In_Txn(id, YES);
    }    
    Setup_Window();
  }                                                  // void Init_Variables(0)

  public void body() {
    // Body for Scheduler. This is the 'behaviour' of the Scheduler.
    boolean stop; 
    Message msg;

    Init_Variables();

    Show_Window();

    Info_Pub("Current Table: "+Table().Name()+"\t\t(total of "
                                                  +Option.Total_Tables()+")");
    Table().Display();


    Info_Pub("\n\t\t <- Start of DCP Simulation ->\n");
  
    Send_FailSafe_Exit();                 // Send VERY delayed message to self

    // Pre-Transaction stage
    // Decide which site is a participant and/or coordinator

    Set_Coord_In_Txn(Option.Coord_Site(), YES);     // Site of 1st coordinator 

    Set_x_Part_CorA(Option.Num_Abort(), ABORT);   // Set num of Parts to abort
    Display_Coord_List();
    Display_Part_List();

    // Send 'Participate' message to all sites indicating their roles in txn 
    Send_Participate_Msgs();


    // Start transaction by sending first message to coordinator
    stop = Start_Txn();              // Will return false if able to start txn

    // Transaction started 
    while (!stop) {
        msg = Receive_Message();
        if (msg.Dest_Type() != SCHED) {
	    // Message not intended for Scheduler so pass it on
            Forward_Message(msg);
        } 
        else {                           // Message for scheduler so handle it
            stop = Handle_Sched_Msg(msg.Clone());
        }
    }

    End_Txn();

    sim_hold(10.0);
    Info("<- Terminated Scheduler: "
                              +Global.Name_CorA(this.Global_Decision())+"\n"); 
    Display_Statistics();
  }                                                            // void body(0)


  private void Send_FailSafe_Exit() {
      // Method to send a TIMEOUT message to the scheduler with an 
      // EXTREMELY long delay so that it will arrive well after the txn
      // has finished in case the sites have blocked.
      Message msg;
      msg = new Message("END-TXN", SCHED, SCHED, null, SCHED, SCHED, -1,
                        UNKNOWN, null, true);

      sim_schedule(port_self, 500.0, SCHED, msg.Clone());
      Trace_Sent(msg);
      Global.Inc_Ctrl_Msgs_Sent();
      Debug("control msgs sent "+Global.Ctrl_Msgs_Sent());
  }                                              // void Send_FailSafe_Exit(0)



  private void Send_Participate_Msgs() {
      // Method to inform sites what their roles are in the txn.
      // COORD - Coordinator, PART - Participant, BOTH - Coord and Part
      int     id, decision, role;        
      Message msg;                         

      role = UNKNOWN;
      for (id=1; id <= Option.Num_Sites(); id++) {
          if (Coord_In_Txn(id)) {
              role = COORD;
          }
          if (Part_In_Txn(id)) {
              // Each site with a participant that is in txn
	      if (role == COORD) {
                  role = BOTH;
              }
              else {
                  role = PART;
              }
          }
          msg = new Message("PARTICIPATE", SITE, id, null, 
                            SCHED, SCHED, -1, role, Part_CorA_List(), true);
          Send_Message(msg);
      }
  }                                           // void Send_Participate_Msgs(0)

  private boolean Start_Txn() {
      // Method to start the txn by sending First Message to Coord and a
      // list of participants.
      boolean stop = true;                      // Return false if started txn
      Msg     act_msg;                                       // action message
      Message msg;
 
      Info("\n-> Starting Transaction\n");

      act_msg = Table().First_Msg_COORD();     // Get 1st msg to send to Coord 
      msg = null;

      if (act_msg != null) {               // Found first message to start txn
          msg = new Message(act_msg.Name(), COORD, Get_Coord_Id(), null,
                                   SCHED, SCHED, -5, NONE, Part_List(), true);
          Send_Message(msg);
          stop = false;
      }
      else {
          Error("Could not find first message to send to Coord");
      }
      return stop;
  }                                                    // boolean Start_Txn(0)

  private void End_Txn() {
      // Method to end the txn by sending END to all sites
      int     id;
      Message msg;
      for (id=1; id <= Option.Num_Sites(); id++) {
          msg = new Message("END-TXN", SITE, id, null, 
                            SCHED, SCHED, -1, Global_Decision(), null, true);
          Send_Message(msg);
      }
  }                                                         // void End_Txn(0)

  private boolean Handle_Sched_Msg(Message msg) {
      // Method to handle any message for the Scheduler
      // Returns true if txn has ended
      boolean stop = false;

      if (msg.Round() == 0) {                        // Transaction terminated
          Debug("Transaction terminated - received "+msg.Name()+" from "
                +Global.Short_Type(msg.Src_Type())+msg.Src_Id());
          Set_Global_Decision(msg.Info());
          stop = true;
      }
      if (msg.Round() == -1) {
          Debug("Transaction terminated - received failsafe "
                +msg.Name()+" from self");
          stop = true;
          Error("<- Terminated TXN - simulation has blocked");
      }
      return stop;
  }                                             // boolean Handle_Sched_Msg(1)

  private static int Num_Coords() {
      // Method to return the current number of coordinators.
      int num;
      num = coord_list[0];
      return num;
  }                                                       // int Num_Coords(0)

  private static int Num_Parts() {
      // Method to return the number of participants in list,
      // number held in index zero.
      int num;

      num = part_list[0];

      return num;
  }                                                        // int Num_Parts(0)

  private static int Num_CorA_Parts() {
      // Method to return the number of participants in commit/abort list,
      // number held in index zero.
      int num = 0;

      num = part_CorA_list[0];

      return num;
  }                                                   // int Num_CorA_Parts(0)

  private static int Num_Commited() {
      // Method to return the number of participants who were to commit.
      int num = 0;
      int id;

      for (id=1; id <= Option.Num_Sites(); id++) {
          if (Part_CorA(id) == COMMIT) {
              num++;
          }
      }
      return num;
  }                                                     // int Num_Commited(0)

  private static int Num_Aborted() {
      // Method to return the number of participants who were to abort.
      int num = 0;
      int id;

      for (id=1; id <= Option.Num_Sites(); id++) {
          if (Part_CorA(id) == ABORT) {
              num++;
          }
      } 
      return num;
  }                                                     // int Num_Commited(0)

   private static void Set_Port(int id, String port_name, String p_gif) {
      // Method to set a port in the port array at a given index (the 
      // site's id. 
      // Changed for new Sim_anim Sim_port constructor
      if ((id <= 0) || (id > Option.Num_Sites())) {
          Error("ID "+id+" invalid for port array");
      }
      port_site[id] = new Sim_port(port_name, p_gif,
                                   Option.Port_Side(SCHED, id, SITE), 
                                   Option.Port_Pos(SCHED, id));
  }                                                        // void Set_Port(3)

  private static Sim_port Get_Port(int id) {
      // Method to return a site's port from the port list given the 
      // site ID.
      if (id <= 0) {
	  // Error checking - no part. has ID=0, so has no port at index 0;
          Error("Port index "+id+" out of range");
      }
      return port_site[id];
  }                                                    // Sim_port Get_Port(1)

  private static int Get_Coord_Id() {
      // returns the id of coordinator when only one in list
      int id, ret_id = 0;
      if (Num_Coords() != 1) {
	  Error("Number of Coords "+Num_Coords()+" requires choice");
      }
      else {
          id = 1;
 	  while ((id <= Option.Num_Sites()) && (ret_id == 0)) {
	      if (Coord_In_Txn(id)) {
                  ret_id = id;
              }
              id++;
          }
      }
      return ret_id;
  }                                                     // int Get_Coord_Id(0)

  private static void Set_Part_In_Txn(int id, int yesno) {
      // Method to set a participant in the part list to be included (YES)
      // excluded (NO) from from being sent messages.  Index is ID.
      if (yesno == YES) {
          if (part_list[id] != YES) {
              part_list[0]++;
          }
          part_list[id] = yesno;
          if ((Part_CorA(id) != ABORT) && (Part_CorA(id) != COMMIT)) {
              Set_Part_CorA(id, COMMIT);                // Set Part. to COMMIT
          }
      }
      else if (yesno == NO) {
          if (part_list[id] == YES) {
              part_list[0]--;
          }
          part_list[id] = yesno;
          Set_Part_CorA(id, UNKNOWN);  
      }
  }                                                 // void Set_Part_In_Txn(2)

  private static boolean Part_In_Txn(int id) {
      // Method to return a participants intended voting decision or 
      // NO if excluded from being sent messages.
      boolean in_txn = false;                                      // Default

      if (id <= 0) {
	  // Error checking - no part. has id 0 - at present.
          Error("Part ID "+id+" out of range");
      }
      else if (part_list[id] == YES) {
          in_txn = true;
      }
      return in_txn;
 }                                                   // boolean Part_In_Txn(1)

 private static int Part_CorA(int id) {
      // Method to return a participants intended voting decision or 
      // NONE if excluded from being sent messages.
      int vote = NONE;                                             // Default

      if (id <= 0) {
	  // Error checking - no part. has id 0 - at present.
          Error("Part ID "+id+" out of range");
      }
      else {
          vote = part_CorA_list[id];
      }
      return vote;
  }                                                        // int Part_CorA(1)

  private static void Set_Part_CorA(int id, int vote) {
      // Method to set a participant in the part list to be included 
      // with an intended vote (COMMIT/ABORT) or excluded (NONE) from
      // from being sent messages.  Index is ID.
      part_CorA_list[id] = vote;

      if (vote == UNKNOWN) {
  	  if (part_CorA_list[id] != UNKNOWN) {
              part_CorA_list[id] = vote;
              part_CorA_list[0]--;
          }
      }
      else {
  	  if ((part_CorA_list[id] != COMMIT) 
                                           || (part_CorA_list[id] != ABORT)) {
              part_CorA_list[id] = vote;
              part_CorA_list[0]++;
          }
          else {
              part_CorA_list[id] = vote;
          } 
      }
  }                                                   // void Set_Part_CorA(2)

  private static void Set_x_Part_CorA(int num, int vote) {
      // Method to set a number of participants to ABORT or whatever the
      // vote is.  At present this selects 'num' parts starting from the
      // end of list.  May consider more even distribution.
      int id, count;

      id = Option.Num_Sites();
      count = 0;
      while ((count != num) && (id >= 1)) {
	  if (Part_In_Txn(id)) {
	      if (Part_CorA(id) != vote) {
                  Set_Part_CorA(id, vote);
              }
              count++;
          }
          id--;
      }
  }                                                 // void Set_x_Part_CorA(2)

  private static void Set_Coord_In_Txn(int id, int yesno) {
      // Method to add a Coordinators ID in coord_list[].  Will increment
      // the number of coords as well.
      int idx = 1;

      if (id <= 0) { 
  	  // Consider error checking 
          Error("Coord ID "+id+" out of range");
      }
      else {
	  if (coord_list[id] == yesno) {
              Debug("Site's Coord status not changed - "+yesno);
          }
          else if (yesno == YES) {
              coord_list[id] = YES;
              coord_list[0]++;
              Debug("Site "+id+" is set as a Coordinator");
          }
          else if (yesno == NO) {
              coord_list[id] = NO;
              coord_list[0]--;
              Debug("Site "+id+" is no longer a Coordinator");
          }
          else {
              Error("Unknown YES/NO status in Set_Coord_In_Txn(2)");
          }
      }
  }                                                // void Set_Coord_In_Txn(2)

  private static boolean Coord_In_Txn(int id) {
      // Method to return true is site is a Coordinator in txn
      boolean reply = false;
      if (id <= 0) {
          Error("Coord id "+id+"  out of range");
      }
      else {
          if (coord_list[id] == YES) {
              reply = true;
          }
      }
      return reply;
  }                                                 // boolean Coord_In_Txn(1)

  private static void Display_Coord_List() {
      // Method to display the coordinators in the Scheduler's coord_list[].
      int id;

      for (id=1; id <= Option.Num_Sites(); id++) {
  	  if (Coord_In_Txn(id)) {
              Debug("\tsite list: Coord"+id+" is in list");
          }
      }
  }                                              // void Display_Coord_List(0)

  private static void Display_Part_List() {
      // Method to display the participants in the Scheduler's part_list[].
      int id, vote;

      for (id=1; id <= Option.Num_Sites(); id++) {
  	  if (Part_In_Txn(id)) {
              vote = Part_CorA(id);
              Debug("\tsite list: Part"+id
                   +" ("+Global.Name_CorA(vote)+") is in list"); 
 	  }
          else if (Part_CorA(id) != UNKNOWN) {
              Debug("Part"+id+" has vote but not in txn");
          }
      }
  }                                               // void Display_Part_List(0)

  private static int[] Part_List() {
      // Method to return the list of participants in txn.
      int[] p_list = new int[MAX_SITES+1];

      if (Num_Parts() > 0) {
          System.arraycopy(part_list, 0, p_list, 0, Option.Num_Sites()+1);
      }
      else {
          Debug("Returning empty part list");
      }
      return p_list;
  }                                                      // int[] Part_List(0)

  private static int[] Part_CorA_List() {
      // Method to return the list of participants and their votes
      int[] CorA_list = new int[MAX_SITES+1];

      if (Num_CorA_Parts() > 0) {
          System.arraycopy(part_CorA_list, 0, 
                           CorA_list, 0, Option.Num_Sites()+1);
      }
      else {
          Debug("Returning empty part C or A list");
      }
      return CorA_list;
  }                                                 // int[] Part_CorA_List(0)

  private static void Inc_Local_Msg_Round() {
      // Method to increment the local message round
      local_msg_round++;
  }                                             // void Inc_Local_Msg_Round(0)

  private static int Local_Msg_Round() {
      // Method to return the local message round
      return local_msg_round;
  }                                                  // int Local_Msg_Round(0)

  private void Set_Global_Decision(int outcome) {
      // Method to set the final outcome of protocol i.e. Commit or Abort
      if (((outcome == COMMIT) || (outcome == ABORT)) 
                                              && (final_outcome == UNKNOWN)) {
          final_outcome = outcome;
      }
      Trace_P(Global.Name_CorA(final_outcome));
  }                                             // void Set_Global_Decision(1)

  private static int Global_Decision() {
      // Method to return the final outcome of protocol i.e. Commit or Abort
      int outcome = UNKNOWN;

      if ((final_outcome == COMMIT) || (final_outcome == ABORT)) {
          outcome = final_outcome;
      }     
      return outcome;
  }                                                  // int Global_Decision(0)

  private static Table Table() {
      // Method to return the current table
      return Option.Table();
  }                                                          // Table Table(0)


  private static void Display_Statistics() {
      // Method to display the message and message round statistics after
      // completion of the transaction.
      // Will enhance later.
      int i, n, num_fail, num_have_failed;
      int num_commit, num_abort, adjusted_timeouts;

      n = Option.Num_Sites();
      Info_Pub("\n\t\t\tSTATISTICS");
      Info_Pub("\t\t\t----------\n");
      Info_Pub("Number of Participants     \t: "+n);
      Info_Pub("Coordinator at site        \t: "+Option.Coord_Site());
      Info_Pub("\nFailure determined by      \t: "
                                   +Global.Name_FR_Method(Option.F_Method()));
      if (Option.Fail(INTERVAL)) {
          Info_Pub("     Interval between failures \t: "
                                            +Option.F_Interval()+" messages");
      }
      else if (Option.Fail(STATES)) {
  	  for (i=1; i < Table().Total_States(PART); i++) {
              num_fail = Option.Num_To_Fail(i);
              num_have_failed = Global.Total_Have_Failed(i);
   	      if (num_fail > 0) {
                  Info_Pub("Failures in state "+i+": "
                           +Table().State_Name(PART, i)
                           +" \t: "+num_fail+" (max) - "
                           +num_have_failed+" (total)");
              }
          }
      }
      else if (Option.Fail(RANDOM)) {
          Pr_Dist.Display_Pr_F_Stats();
      }

      Info_Pub("\nRecovery determined by     \t: "
                                   +Global.Name_FR_Method(Option.R_Method()));

      if (Option.Recover(INTERVAL)) {
          Info_Pub("     Interval between recovery \t: "
                                            +Option.R_Interval()+" messages");
      }
      if (Option.Recover(RANDOM)) {
          Pr_Dist.Display_Pr_R_Stats();
      }

      Info_Pub("\n");

      Table().Display_Stats(n);

      Info_Pub("\nActual statistics for "+Table().Name()+" to "
           +Global.Name_CorA(Global_Decision())+" -");
      Info_Pub("\tTotal messages sent   \t: "+Global.Messages_Sent());
      Info_Pub("\tTotal messages rec    \t: "+Global.Messages_Rec());
      Info_Pub("\tControl messages sent \t: "+Global.Ctrl_Msgs_Sent());
      Info_Pub("\tControl messages rec  \t: "+Global.Ctrl_Msgs_Rec());
      Info_Pub("\n" +
               "\tReal messages sent    \t: "+Global.Real_Msgs_Sent());
      Info_Pub("\tReal messages received\t: "+Global.Real_Msgs_Rec());
      Info_Pub("\n" +
               "\tTotal message rounds  \t: "+Global.Message_Round());
      Info_Pub("\n" + 
               "\tTotal DT log writes   \t: "+Global.DT_Log_Writes());
      Info_Pub("\tTotal DT log reads    \t: "+Global.DT_Log_Reads());
      num_commit = Num_Commited();
      num_abort = Num_Aborted();
      Info_Pub("\n" + 
               "\tWilling to commit txn \t: "+num_commit);
      Info_Pub("\tWilling to abort txn  \t: "+num_abort);
      if ((num_commit+num_abort) != (Option.Num_Sites())) {
          Error_Pub("Not all Participants accounted for");
      } 
      Info_Pub("\n" +
               "\tParticipant failures \t: "+Global.Failed_PART());
      Info_Pub("\tParticipant recovery \t: "+Global.Recovered_PART());
      if (Global.Timeout_Cycles() > 0) {
          adjusted_timeouts = (Global.Timeouts_COORD()  
                    - (Global.Timeout_Cycles() * (DETECT_MESSAGE_LOOP - 1)) );
          Info_Pub("\n" +
               "\tCoordinator Timeouts  \t: "+Global.Timeouts_COORD()
              +" ("+adjusted_timeouts+" adj.)");
      }
      else {
          Info_Pub("\n" +
               "\tCoordinator Timeouts  \t: "+Global.Timeouts_COORD());
      }

      Info_Pub("\tTimeout loops detected\t: "+Global.Timeout_Cycles()+"\n");
      Info_Pub("\tFailure loops detected\t: "+Global.FailR_Cycles()+"\n");

      if (Global.Num_Errors() > 0) {
          Scheduler.Info_Pub("\t\t     "+Global.Num_Errors()
                             +" errors during txn");
      }
      Scheduler.Info_Pub("\n\t\t <- End of DCP Simulation ->\n");

  }                                              // void Display_Statistics(0)

  private Message Receive_Message() {
      // Method to wait for a message (event) or return the first queued
      // in buffer.  The buffered message will be returned first.
      Message   msg;
      Sim_event ev = new Sim_event();;

      Debug("number of events queued: "+sim_waiting());
      Debug("waiting for event to arrive");
      sim_get_next(ev);

      msg = (Message) ev.get_data();
      Trace_Received(msg);
      if (msg.Round() == Local_Msg_Round() + 1) {
          Inc_Local_Msg_Round();
      }

      if (msg.Dest_Type() == SCHED) {
	  // Final destination of message is Scheduler so increment count.
          Global.Inc_Ctrl_Msgs_Rec();
          Debug("control msgs received "+Global.Ctrl_Msgs_Rec());
          Debug("messages received "+Global.Messages_Rec());
      }
      return msg;
  }                                              // Message Receive_message(0)

  private void Forward_Message(Message snd_msg) {
      // Method to forward a given message to the correct dest
      Message msg;

      msg = snd_msg.Clone();

      Send_Message(msg);
  }                                                 // void Forward_Message(1)

  private void Send_Message(Message snd_msg) {
      // Method to send a given message
      Message msg;
 
      msg = snd_msg.Clone();

      sim_schedule(Get_Port(msg.Dest_Id()), LINK_DELAY, SITE, msg);
      Trace_Sent(msg);
      if (msg.Src_Type() == SCHED) {
	  // Message originates at Scheduler, not just passed on from site,
	  // so increment messages sent. 
  	  if (!msg.Control()) {
              Info("Sent "+msg.Name());
              Global.Inc_Real_Msgs_Sent();
              Debug("real msgs sent "+Global.Real_Msgs_Sent());
   	  } 
          else {
              Global.Inc_Ctrl_Msgs_Sent();
              Debug("control msgs sent "+Global.Ctrl_Msgs_Sent());
          }
      }
  }                                                    // void Send_Message(1)

  private void Trace_Sent(Message msg) {
      // Method to output trace that message was sent to entity  
      String port_name = null;

      if (msg.Src_Type() == SCHED) {
          Trace("sent "+msg.Sent_Data(SCHED));
          Debug("sent "+msg.Sent_Data(SCHED));
      }
      else {
          Trace("(sent) forwarding "+msg.Sent_Data(SCHED));
          Route("(sent) forwarding "+msg.Sent_Data(SCHED));
      }
      if (msg.Dest_Type() == SCHED) {
          port_name = new String(port_self.get_pname());
      }
      else {
  	  port_name = new String(Get_Port(msg.Dest_Id()).get_pname()); 
      }
      // Added for SimAnim
      if (msg.Control()) {
  	 if (Option.Show_Ctrl_Messages()) {   
             Trace_S(port_name, msg.Name().toLowerCase());
         }
      }
      else{
         Trace_S(port_name, msg.Name().toUpperCase());
      }
  }                                                      // void Trace_Sent(1)

  private void Trace_Received(Message msg) {
      // Method to output trace that message was sent to entity, if required  
      if (msg.Dest_Type() == SCHED) {
          Trace("received "+msg.Rec_Data(SCHED));
          Debug("received "+msg.Rec_Data(SCHED));
      }
      else {
          Trace("(rec) forwarding "+msg.Rec_Data(SCHED));
          Route("(rec) forwarding "+msg.Rec_Data(SCHED));
      }
  }                                                  // void Trace_Received(1)

  private static void Setup_Window() {
      // Method to open the scheduler's window on screen 
      if (Option.Display_GUI()) {
          Debug("Setting up GUI window");
          sched_window = new W_Box("Monitor");
      } 
  }                                                     // void Open_Window(0)

  protected static void Reset_Window() {
      // Method to reset the window without killing it
      if (sched_window != null) {
          Hide_Window();
          sched_window.Clear();      
      }
  }                                                    // void Reset_Window(0) 

  private static void Show_Window() {
      Debug("Showing GUI window");
      if (Option.Display_GUI()) {
          Global.Window_Visible(sched_window, true);
      }
  }                                                     // void Show_Window(0)

  private static void Hide_Window() {
      Debug("Hiding GUI window");
      Global.Window_Visible(sched_window, false);
  }                                                     // void Hide_Window(0)

  private static void Close_Window() {
      // Method to close the scheduler's window on screen
      if (sched_window != null) {
          sched_window.dispose();
      } 
  }                                                    // void Close_Window(0)

  private static void Debug(String note) {
      // Method to output debugging comments
      Global.Debug_CLI("Sched : "+note);
      Debug_GUI(note);
  }                                                           // void Debug(1)

  protected static void Debug_Pub(String note) {
      // Method to output debugging comments
      Global.Debug_CLI(note);
      Debug_GUI(note);
  }                                                       // void Debug_Pub(1)

  private static void Debug_GUI(String note) {
      // Method to output debugging comments via graphical user interface
      if ((Option.Show_Debug()) && (Option.Display_GUI())) {
	  if (sched_window != null) {
              sched_window.Add_Output_Entry(note);  
          }
      }
  }                                                       // void Debug_GUI(1)

  private static void Error(String note) {
      // Method to output error messages
      Global.Error_CLI("Sched : "+note);
      Error_GUI(note);
  }                                                           // void Error(1)

  protected static void Error_Pub(String note) {
      // Method to output error messages
      Global.Error_CLI(note);
      Error_GUI(note);
  }                                                       // void Error_Pub(1)

  private static void Error_GUI(String note) {
      // Method to output error messages via graphical user interface
      if ((Option.Show_Error()) && (Option.Display_GUI())) {
	  if (sched_window != null) {
              sched_window.Add_Error_Entry(note);  
          }
      }
  }                                                       // void Error_GUI(1)

  private static void Info(String note) {
      // Method to output information if required
      Global.Info_CLI("Sched : "+note);
      Info_GUI(note);
  }                                                            // void Info(1)

  protected static void Info_Pub(String note) {
      // Method to output information if required
      Global.Info_CLI(note);
      Info_GUI(note);
  }                                                        // void Info_Pub(1)

  private static void Info_GUI(String note) {
      // Method to output info via graphical user interface
      if ((Option.Show_Info()) && (Option.Display_GUI())) {
	  if (sched_window != null) {
              sched_window.Add_Output_Entry(note);  
          }
      }
  }                                                        // void Info_GUI(1)

  private static void Route(String note) {
      // Method to output route debugging comments
      if (Option.Show_Routing()) {
          Debug(note);
      }
  }                                                           // void Route(1) 

  private void Trace_S(String port, String note) {
      // Method to output sending of event traces in tracefile
      // port : name of port event sent from
      // note : name of event

      if ((SIMANIM) && (Option.Show_Messages())) {  
  	  // Only for SimAnim version
          sim_trace(1, "S "+port+" "+note);
      }
  }                                                         // void Trace_S(2)

  private void Trace_P(String param2) {
      // Method to output entity parameter traces in tracefile
      // param1 : global decision

      if (SIMANIM) {
  	  // Only for SimAnim version
          sim_trace(1, "P A "+param2);
      }
  }                                                         // void Trace_P(1)

  private void Trace(String note) {
      // Method to output trace in tracefile
      if ((!SIMANIM) && (Option.Show_Trace())) {
  	  // Must not add extra traces if SimAnim version
          sim_trace(1, "Sched : "+note);
      }
  }                                                           // void Error(1)

}                                                           // class Scheduler

/*---------------------------- End of Scheduler.java ------------------------*/

