
/*
** Author : Steven Ramsay <smr@dcs.ed.ac.uk>, 08/01/97
** Files  : Table.java, Table.class
** Package: simdcp
** Update : 11/06/97 
**
** Description:
** Table structure for representing the protocols.
*/

package eduni.simdcp;

/* Methods defined in abstract class Table
** PUBLIC
** final  void    Init_Table(String)
** final  void    Add_States()
** final  String  Ctrl_Name(int)
** final  void    Set_State_Name(int, int, String)
** final  String  State_Name(int, int)
** final  void    Set_Message_Name(int, String)
** final  String  Message_Name(int)
** final  void    Set_Log_Name(int, String)
** final  String  Log_Name(int)
** final  void    Set_Recover_State(int, int)
** final  int     Recover_State(int)
** final  boolean Fail_Possible(int, int)
** final  void    Display_Stats(int)
** static void    Error(String)
** static void    Info(String)
**
** ABSTRACT
** protected void Display_Stats(int)
** private   void Setup_Table()
** private   void Add_Actions()
**
** PROTECTED
** final  String  Name()
** final  int     Total_States(int)
** final  State   State(int, int)
** final  Msg     First_Msg_COORD()
** final  void    Display()
**
** PRIVATE
**        void    Add_State(int, State)
**        void    Set_Control_Names()
*/

abstract public class Table implements Constants {
  // This class is the representation of CCS in tabular form.  Each protocol 
  // will have its own table.  Each table consists of a set of states for the
  // coordinator algorithm and a set for the participant algorithm.  
  private String   table_name;                            // name of the table
  private int      COORD_total_states;         // total coord states available 
  private int      PART_total_states;           // total part states available
  private States[] COORD_states;                            // the coord sites
  private States[] PART_states;                             // the part states 

  private String[] control_names;               // Control names used in table
  private String[] COORD_state_names;         // Following defined in subclass
  private String[] PART_state_names;
  private String[] message_names;
  private String[] log_names;

  private boolean[] COORD_fail_in_state;
  private boolean[] PART_fail_in_state;

  private int COORD_recover_state;
  private int PART_recover_state;

  private   abstract void Setup_Table();
  private   abstract void Add_Actions();
  protected abstract void Display_Stats(int n);

  public final void Init_Table(String name) {
      this.table_name         = name;
      this.COORD_states       = new States[MAX_STATES + 1];
      this.PART_states        = new States[MAX_STATES + 1];
      this.control_names      = new String[MAX_CTRL_NAMES + 1];
      this.COORD_state_names  = new String[MAX_STATES + 1];
      this.PART_state_names   = new String[MAX_STATES + 1];
      this.message_names      = new String[MAX_MSG_NAMES + 1];
      this.log_names          = new String[MAX_LOG_NAMES + 1];
      this.COORD_fail_in_state= new boolean[MAX_STATES + 1];
      this.PART_fail_in_state = new boolean[MAX_STATES + 1];
      this.COORD_total_states = 0;
      this.PART_total_states  = 0;
      this.Set_Control_Names();
  }                                                      // void Init_Table(1)

  protected final String Name() { 
      return this.table_name;
  }                                                          // String Name(0)

  protected final int Total_States(int type) {
      int total = 0;

      if (type == COORD) {
          total = COORD_total_states;  
      }
      else if (type == PART) {
          total = PART_total_states;  
      }
      else {
          Error("Unknown type in Total_States(1)");
      }
      return total;
  }                                                     // int Total_States(1)

  public final void Add_States() {
      int i;

      for (i=1; i <= MAX_STATES; i++) {
          if (COORD_state_names[i] != null) {
              Add_State(COORD, new States(i, COORD_state_names[i]));
          }
          if (PART_state_names[i] != null) {
              Add_State(PART, new States(i, PART_state_names[i]));
          }
      } 
  }                                                      // void Add_States(0)

  private void Add_State(int type, States new_state) {
      if (type == COORD) {
          if (this.COORD_total_states < MAX_STATES) {
              this.COORD_total_states++;
              this.COORD_states[this.COORD_total_states] = new_state;
          }
          else {
              Error("Cannot add another coord state.");
          }
      }
      else if (type == PART) {
          if (this.PART_total_states < MAX_STATES) {
              this.PART_total_states++;
              this.PART_states[this.PART_total_states] = new_state;
          }
          else {
              Error("Cannot add another part state.");
          }
      }
      else {
          Error("Unknown type in Add_State(2)");
      }
  }                                                       // void Add_State(2)

  protected final States State(int type, int number) {
      States tmp_state = null;

      if (type == COORD) {
          if ((number > 0) && (number <= this.Total_States(COORD))) {
              tmp_state = this.COORD_states[number];
          }
          else {
              Error("No coord state exists");
          }
      }
      else if (type == PART) {
          if ((number > 0) && (number <= this.Total_States(PART))) {
              tmp_state = this.PART_states[number];
          }
          else {
              Error("No part state exists");
          }
      }
      else {
          Error("Unknown type in State(2)");
      }
      return tmp_state;
  }                                                         // States State(2)

  private void Set_Control_Names() {
      control_names = new String[MAX_CTRL_NAMES+1];

      control_names[1] = new String("COMMIT");
      control_names[2] = new String("ABORT");
      control_names[3] = new String("VOTE");
      control_names[4] = new String("END");
      control_names[5] = new String("TIMEOUT");
  }                                               // void Set_Control_Names(0)

  public final String Ctrl_Name(int num) {
      String ctrl_name = null;

      if ((num < 1) || (num > MAX_CTRL_NAMES)) {
          Error("Control name index out of range"); 
      }
      else {
          ctrl_name = this.control_names[num];
      }
      return ctrl_name;
  }                                                     // String Ctrl_Name(1)

  public final void Set_State_Name(int type, int num, String name, 
                                                        boolean possible) {
      if ((num < 1) || (num > MAX_STATES)) {
          Error("State number out of range: "+num);
      }
      else {
  	  if (type == COORD) {
              this.COORD_state_names[num] = new String(name); 
              this.COORD_fail_in_state[num] = possible;
	  }
	  else if (type == PART) {
              this.PART_state_names[num] = new String(name);
              this.PART_fail_in_state[num] = possible;
          }
          else {
              Error("Unknown type in Set_State_Name(3)");
          }
      }   
  }                                                  // void Set_State_Name(3)

  public final String State_Name(int type, int num) {
      String name = new String("_ERROR_");
      if ((num < 1) || (num > this.Total_States(type))) {
          Error("State number out of range: "+num);
      }
      else {
  	  if (type == COORD) {
              name = this.COORD_state_names[num]; 
	  }
	  else if (type == PART) {
              name = this.PART_state_names[num];
          }
          else {
              Error("Unknown type in State_Name(2)");
          }
      }
      return name;
  }                                                    // String State_Name(2)

  public final void Set_Message_Name(int num, String name) {
      if ((num < 1) || (num > MAX_MSG_NAMES)) {
          Error("Message number out of range: "+num);
      }
      else {
          this.message_names[num] = new String(name); 
      }
  }                                                // void Set_Message_Name(2)

  public final String Message_Name(int num) {
      String msg_name = new String("_ERROR_");

      if ((num < 1) || (num > MAX_MSG_NAMES)) {
          Error("Message name index out of range"); 
      }
      else {
          msg_name = this.message_names[num];
      }
      return msg_name;
  }                                                  // String Message_Name(1)

  public final void Set_Log_Name(int num, String name) {
      if ((num < 1) || (num > MAX_LOG_NAMES)) {
          Error("Log number out of range: "+num);
      }
      else {
          this.log_names[num] = new String(name); 
      }
      this.log_names[0] = new String("");
  }                                                    // void Set_Log_Name(2)

  public final String Log_Name(int num) {
      String log_name = new String("_ERROR_");

      if ((num < 0) || (num > MAX_LOG_NAMES)) {
          Error("Log name index out of range"); 
      }
      else {
          log_name = this.log_names[num];
      }
      return log_name;
  }                                                      // String Log_Name(1)

  public final void Set_Recover_State(int type, int num) {
      // Set the number of the state in which to recover
      if ((num < 1) || (num > this.Total_States(type))) {
          Error("Recover state number invalid");
      }
      else {
          if (type == COORD) {
              COORD_recover_state = num;
          }
          if (type == PART) {
              PART_recover_state = num;
          }
      }
  }                                               // void Set_Recover_State(2)

  public final int Recover_State(int type) {
      // Set the number of the state in which to recover
      int num = 0;

      if (type == COORD) {
          num = COORD_recover_state;
      }
      else if (type == PART) {
          num = PART_recover_state;
      }
      return num;
  }                                                    // int Recover_State(1)

  public final boolean Fail_Possible(int type, int num) {
      // Returns true if it is possible to fail in this state 
      boolean possible = false;

      if (type == COORD) {
          possible = COORD_fail_in_state[num];
      }
      else if (type == PART) {
          possible = PART_fail_in_state[num];
      }
      return possible;
  }                                                    // int Fail_Possible(1)



  protected final Msg First_Msg_COORD() {
      // Method to find the first message which has to be sent to coord to
      // start transaction.
      // Assumes : Coord starts message passing and that first message
      // received by Coord (before it sends one) is in state 1.
      int     act_num, state_num, choice_num;
      boolean stop;
      States  state;
      Choices choice;
      Actable action;
      Msg     msg = null;
    
      act_num = 1;
      state_num = 1;
      choice_num = 1;
      stop = false;

      while (!stop) {
          state = this.State(COORD, state_num);
          choice_num = 1;
          while ((state.Total_Choices() >= 1) &&
                           (choice_num <= state.Total_Choices()) && (!stop)) {

              choice = state.Choice(choice_num);
              act_num = 1;
              while ((choice.Total_Actions() >= 1) && 
                             (act_num <= choice.Total_Actions()) && (!stop)) {

                  action = choice.Action(act_num);
                  if (action instanceof Msg) {
                      msg = (Msg) action;
                      if ((msg.Receive()) && (msg.Round() == 0)) {
                          stop = true;
                      }
                      else {
                          Error("1st message not a receive");
                          stop = true;
                          msg = null;
                      }
                  }
                  else {
                      act_num++;
                  }
              }
              if (!stop) {
                  choice_num++;
              }
	  }
          if (!stop) {
              Error("1st Message not found in first state");
          }
      }
      return msg;
  }                                                  // Msg First_Msg_COORD(0)

  protected final void Display() {
      int    i;
      States tmp_state;

      Info("\nTable: "+this.Name());
      Info("\n\t- Coordinator States -");
      for (i=1; i <= this.Total_States(COORD); i++) {
           tmp_state = State(COORD, i);
           if (tmp_state != null) {
               Info("S"+i+"-"+tmp_state.Name()+":");
               tmp_state.Display();
           }
      }
      Info("\n\t- Participant States -");
      for (i=1; i <= this.Total_States(PART); i++) {
           tmp_state = State(PART, i);
           if (tmp_state != null) {
               Info("S"+i+"-"+tmp_state.Name()+":");
               tmp_state.Display();
           }
      }
      Info("End of Table\n");
  }                                                         // void Display(0)

  public final void Display_Statistics(int n) {
      // Method to output the expected statistics for protocol based on 
      // n participants
      Info("Expected statistics for "+this.Name()+" to COMMIT -");
  }                                              // void Display_Statistics(1)

  public static final void Error(String note) {
      // Method to output error message if required
      Scheduler.Error_Pub(note+" (class Table)");
  }                                                           // void Error(1)

  public static final void Info(String note) {
      // Method to output information if required
      Scheduler.Info_Pub(note);
  }                                                            // void Info(1)

}                                                               // class Table

/*------------------------------ End of Table.java --------------------------*/

