
/*
** Author : Steven Ramsay <smr@dcs.ed.ac.uk>, 21/01/97
** Files  : Table_3PC.java, Table_3PC.class
** Package: simdcp
** Update : 11/06/97 
**
** Description: A specific protocol represented in the table structure.
** Name       : Basic 3-Phase Commit Protocol
** Comments   : Does not handle Coordinator Elections
*/

package eduni.simdcp;

/* Methods defined in final class Table_3PC
** PUBLIC
** Table_3PC(String)
**
** PROTECTED
**        void    Display_Stats(int)
**
** PRIVATE
**        void    SetUp_Table()
**        void    Set_State_Names()
**        void    Set_Message_Names()
**        void    Set_Log_Names()
**        void    Add_Actions()
*/
final class Table_3PC extends Table implements Constants {

  public Table_3PC(String name) {
      super.Init_Table(name);                  // Must be first in constructor
      this.Setup_Table();
  }                                                // constructor Table_3PC(1)

  private void Setup_Table() {
      Info("Constructing table: "+super.Name()+"...");
      this.Set_State_Names();
      this.Set_Message_Names();
      this.Set_Log_Names();
      super.Add_States();
      super.Set_Recover_State(COORD, 7);
      super.Set_Recover_State(PART, 7);
      this.Add_Actions();
      Info("...Finished "+super.Name());
  }                                                     // void Setup_Table(0)
      			
  private void Set_State_Names() {
      Set_State_Name(COORD, 1, "INITIAL",    false);
      Set_State_Name(COORD, 2, "WAIT",       false);
      Set_State_Name(COORD, 3, "PRE-COMMIT", false);
      Set_State_Name(COORD, 4, "COMMIT",     false);
      Set_State_Name(COORD, 5, "ABORT",      false);
      Set_State_Name(COORD, 6, "FINAL",      false);
      Set_State_Name(COORD, 7, "RECOVER",    false);

      Set_State_Name(PART, 1, "INITIAL",    true );
      Set_State_Name(PART, 2, "READY",      true );
      Set_State_Name(PART, 3, "PRE-COMMIT", true );
      Set_State_Name(PART, 4, "COMMIT",     true );
      Set_State_Name(PART, 5, "ABORT",      true );
      Set_State_Name(PART, 6, "FINAL",      false);
      Set_State_Name(PART, 7, "RECOVER",    false);
      Set_State_Name(PART, 8, "ELECTION",   false);
  }                                                 // void Set_State_Names(0)

  private void Set_Message_Names() {
      Set_Message_Name(1, "START");
      Set_Message_Name(2, "PREPARE");
      Set_Message_Name(3, "VOTE-COMMIT");
      Set_Message_Name(4, "VOTE-ABORT");
      Set_Message_Name(5, "PREPARE-TO-COMMIT");
      Set_Message_Name(6, "READY-TO-COMMIT");
      Set_Message_Name(7, "GLOBAL-COMMIT");
      Set_Message_Name(8, "GLOBAL-ABORT");
      Set_Message_Name(9, "ACK");
      Set_Message_Name(10, "HELP-ME");
      Set_Message_Name(11, "ELECT");
  }                                               // void Set_Message_Names(0)

  private void Set_Log_Names() {
      Set_Log_Name(1, "start-3PC");
      Set_Log_Name(2, "ready");
      Set_Log_Name(3, "commit");
      Set_Log_Name(4, "abort");
      Set_Log_Name(5, "end_of_transaction");
  }                                                   // void Set_Log_Names(0)

  private void Add_Actions() {
      States  tmp_state;
      Choices tmp_choice;

      // Coordinator 
      tmp_state = State(COORD, 1);                                  // State 1
      tmp_state.Add_x_Choices(1); 
      tmp_choice = tmp_state.Choice(1);                   // State 1, Choice 1
      tmp_choice.Add_Action(new Msg(Message_Name(1), 0, REC, UNI, NONE));
      tmp_choice.Add_Action(new Log(Log_Name(1), W, NONE));
      tmp_choice.Add_Action(new Msg(Message_Name(2), 1, SND, MLT, NONE));
      tmp_choice.Set_Next_State(2);

      tmp_state = State(COORD, 2);                                  // State 2
      tmp_state.Add_x_Choices(5);
      tmp_choice = tmp_state.Choice(1);                   // State 2, Choice 1 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(4), COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(5), 3, SND, MLT, COMMIT));
      tmp_choice.Set_Next_State(3);
      tmp_choice = tmp_state.Choice(2);                   // State 2, Choice 2 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(4), ABORT));
      tmp_choice.Add_Action(new Log(Log_Name(4), W, ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(8), 3, SND, MLT, ABORT));
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(3);                   // State 2, Choice 3 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(5), 2, REC, UNI, NONE));
      tmp_choice.Add_Action(new Log(Log_Name(4), W, ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(8), 3, SND, MLT, ABORT));
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(4);                   // State 2, Choice 4 
      tmp_choice.Add_Action(new Msg(Message_Name(3), 2, REC, UNI, COMMIT));
      tmp_choice.Set_Next_State(2);      
      tmp_choice = tmp_state.Choice(5);                   // State 2, Choice 5 
      tmp_choice.Add_Action(new Msg(Message_Name(4), 2, REC, UNI, ABORT));
      tmp_choice.Set_Next_State(2);

      tmp_state = State(COORD, 3);                                  // State 3
      tmp_state.Add_x_Choices(3);
      tmp_choice = tmp_state.Choice(1);                   // State 3, Choice 1 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(4), COMMIT));
      tmp_choice.Add_Action(new Log(Log_Name(3), W, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(7), 5, SND, MLT, COMMIT));
      tmp_choice.Set_Next_State(4);
      tmp_choice = tmp_state.Choice(2);                   // State 3, Choice 2 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(5), 6, REC, UNI, COMMIT));
      tmp_choice.Add_Action(new Log(Log_Name(3), W, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(7), 5, SND, MLT, COMMIT));
      tmp_choice.Set_Next_State(4);      
      tmp_choice = tmp_state.Choice(3);                   // State 3, Choice 3 
      tmp_choice.Add_Action(new Msg(Message_Name(6), 4, REC, UNI, COMMIT));
      tmp_choice.Set_Next_State(3);      

      tmp_state = State(COORD, 4);                                  // State 4
      tmp_state.Add_x_Choices(4);
      tmp_choice = tmp_state.Choice(1);                   // State 4, Choice 1 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(4), COMMIT));
      tmp_choice.Add_Action(new Log(Log_Name(5), W, COMMIT));
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(1), 0, SND, UNI, COMMIT));
      tmp_choice.Set_Next_State(6);      
      tmp_choice = tmp_state.Choice(2);                   // State 4, Choice 2 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(5), 6, REC, UNI, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(7), 5, SND, UNI, COMMIT));
      tmp_choice.Set_Next_State(4);
      tmp_choice = tmp_state.Choice(3);                   // State 4, Choice 3 
      tmp_choice.Add_Action(new Msg(Message_Name(9), 6, REC, UNI, COMMIT));
      tmp_choice.Set_Next_State(4);      
      tmp_choice = tmp_state.Choice(4);                   // State 4, Choice 4 
      tmp_choice.Add_Action(new Msg(Message_Name(6), 4, REC, UNI, COMMIT));
      tmp_choice.Set_Next_State(4);      

      tmp_state = State(COORD, 5);                                  // State 5
      tmp_state.Add_x_Choices(5);
      tmp_choice = tmp_state.Choice(1);                   // State 5, Choice 1
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(4), ABORT));
      tmp_choice.Add_Action(new Log(Log_Name(5), W, ABORT));
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(2), 0, SND, UNI, ABORT));
      tmp_choice.Set_Next_State(6);
      tmp_choice = tmp_state.Choice(2);                   // State 5, Choice 2 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(5), 6, REC, UNI, ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(8), 3, SND, UNI, ABORT));
      tmp_choice.Set_Next_State(5);      
      tmp_choice = tmp_state.Choice(3);                   // State 5, Choice 3 
      tmp_choice.Add_Action(new Msg(Message_Name(9), 4, REC, UNI, ABORT));
      tmp_choice.Set_Next_State(5);      
      tmp_choice = tmp_state.Choice(4);                   // State 5, Choice 4 
      tmp_choice.Add_Action(new Msg(Message_Name(3), 2, REC, UNI, COMMIT));
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(5);                   // State 5, Choice 5 
      tmp_choice.Add_Action(new Msg(Message_Name(4), 2, REC, UNI, ABORT));
      tmp_choice.Set_Next_State(5);

      tmp_state = State(COORD, 6);                                  // State 6
      tmp_state.Add_x_Choices(1);
      tmp_choice = tmp_state.Choice(1);                   // State 6, Choice 1 
      tmp_choice.Set_Next_State(0);

      tmp_state = State(COORD, 7);                                  // State 7
      tmp_state.Add_x_Choices(5);
      tmp_choice = tmp_state.Choice(1);                   // State 7, Choice 1 
      tmp_choice.Add_Action(new Log(Log_Name(1), R, NONE));
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, SND, MLT, NONE));   
      tmp_choice.Set_Next_State(2);
      tmp_choice = tmp_state.Choice(2);                   // State 7, Choice 2 
      tmp_choice.Add_Action(new Log(Log_Name(3), R, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, SND, MLT, NONE));   
      tmp_choice.Set_Next_State(2);
      tmp_choice = tmp_state.Choice(3);                   // State 7, Choice 3 
      tmp_choice.Add_Action(new Log(Log_Name(4), R, ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, SND, MLT, ABORT));
      tmp_choice.Set_Next_State(2);
      tmp_choice = tmp_state.Choice(4);                   // State 7, Choice 4 
      tmp_choice.Add_Action(new Log(Log_Name(5), R, COMMIT));
      tmp_choice.Set_Next_State(6);
      tmp_choice = tmp_state.Choice(5);                   // State 7, Choice 5 
      tmp_choice.Add_Action(new Log(Log_Name(5), R, ABORT));
      tmp_choice.Set_Next_State(6);

      // Participant
      tmp_state = State(PART, 1);                                   // State 1
      tmp_state.Add_x_Choices(3);
      tmp_choice = tmp_state.Choice(1);                   // State 1, Choice 1 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(5), 1, REC, UNI, NONE));
      tmp_choice.Add_Action(new Log(Log_Name(4), W, ABORT));
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(2);                   // State 1, Choice 2 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(3), COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(2), 1, REC, UNI, NONE));
      tmp_choice.Add_Action(new Log(Log_Name(2), W, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(3), 2, SND, UNI, COMMIT));
      tmp_choice.Set_Next_State(2);
      tmp_choice = tmp_state.Choice(3);                   // State 1, Choice 3 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(3), ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(2), 1, REC, UNI, NONE));
      tmp_choice.Add_Action(new Log(Log_Name(4), W, ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(4), 2, SND, UNI, ABORT));
      tmp_choice.Set_Next_State(5);

      tmp_state = State(PART, 2);                                   // State 2
      tmp_state.Add_x_Choices(4);
      tmp_choice = tmp_state.Choice(1);                   // State 2, Choice 1 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(5), 3, REC, UNI, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(11), 4, SND, MLT, NONE));
      tmp_choice.Set_Next_State(8);
      tmp_choice = tmp_state.Choice(2);                   // State 2, Choice 2 
      tmp_choice.Add_Action(new Msg(Message_Name(5), 3, REC, UNI, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(6), 4, SND, UNI, COMMIT));
      tmp_choice.Set_Next_State(3);
      tmp_choice = tmp_state.Choice(3);                   // State 2, Choice 3
      tmp_choice.Add_Action(new Msg(Message_Name(8), 3, REC, UNI, ABORT));
      tmp_choice.Add_Action(new Log(Log_Name(4), W, ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(9), 4, SND, UNI, ABORT));
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(4);                   // State 2, Choice 4 
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, REC, UNI, NONE));
      tmp_choice.Set_Next_State(2);

      tmp_state = State(PART, 3);                                   // State 3
      tmp_state.Add_x_Choices(3); 
      tmp_choice = tmp_state.Choice(1);                   // State 3, Choice 1 
      tmp_choice.Add_Action(new Msg(Message_Name(7), 5, REC, UNI, COMMIT));
      tmp_choice.Add_Action(new Log(Log_Name(3), W, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(9), 6, SND, UNI, COMMIT));
      tmp_choice.Set_Next_State(4);
      tmp_choice = tmp_state.Choice(2);                   // State 3, Choice 2 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(5), 5, REC, UNI, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(11), 6, SND, MLT, NONE));
      tmp_choice.Set_Next_State(8);
      tmp_choice = tmp_state.Choice(3);                   // State 2, Choice 3 
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, REC, UNI, NONE));
      tmp_choice.Set_Next_State(3);

      tmp_state = State(PART, 4);                                   // State 4
      tmp_state.Add_x_Choices(4);
      tmp_choice = tmp_state.Choice(1);                   // State 4, Choice 1 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(1), 0, REC, UNI, COMMIT));
      tmp_choice.Set_Next_State(6);
      tmp_choice = tmp_state.Choice(2);                   // State 4, Choice 2 
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, REC, UNI, NONE));
      tmp_choice.Add_Action(new Msg(Message_Name(7), 6, SND, UNI, COMMIT));
      tmp_choice.Set_Next_State(4);
      tmp_choice = tmp_state.Choice(3);                   // State 4, Choice 3 
      tmp_choice.Add_Action(new Msg(Message_Name(7), 6, REC, UNI, COMMIT));
      tmp_choice.Set_Next_State(4);
      tmp_choice = tmp_state.Choice(4);                   // State 4, Choice 4 
      tmp_choice.Add_Action(new Msg(Message_Name(7), 5, REC, UNI, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(9), 6, SND, UNI, COMMIT));
      tmp_choice.Set_Next_State(4);

      tmp_state = State(PART, 5);                                   // State 5
      tmp_state.Add_x_Choices(4);
      tmp_choice = tmp_state.Choice(1);                   // State 5, Choice 1 
      tmp_choice.Add_Action(new Ctrl(Ctrl_Name(2), 0, REC, UNI, ABORT));
      tmp_choice.Set_Next_State(6);
      tmp_choice = tmp_state.Choice(2);                   // State 5, Choice 2 
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, REC, UNI, NONE));
      tmp_choice.Add_Action(new Msg(Message_Name(8), 4, SND, UNI, ABORT));
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(3);                   // State 5, Choice 3 
      tmp_choice.Add_Action(new Msg(Message_Name(8), 4, REC, UNI, ABORT));
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(4);                   // State 5, Choice 4 
      tmp_choice.Add_Action(new Msg(Message_Name(8), 3, REC, UNI, ABORT));
      tmp_choice.Add_Action(new Msg(Message_Name(9), 4, SND, UNI, ABORT));
      tmp_choice.Set_Next_State(5);

      tmp_state = State(PART, 6);                                   // State 6
      tmp_state.Add_x_Choices(1);
      tmp_choice = tmp_state.Choice(1);                   // State 6, Choice 1 
      tmp_choice.Set_Next_State(0);

      tmp_state = State(PART, 7);                                   // State 7
      tmp_state.Add_x_Choices(4);
      tmp_choice = tmp_state.Choice(1);                   // State 7, Choice 1 
      tmp_choice.Add_Action(new Log(Log_Name(0), R, NONE));
      tmp_choice.Add_Action(new Log(Log_Name(4), W, ABORT));   
      tmp_choice.Set_Next_State(5);
      tmp_choice = tmp_state.Choice(2);                   // State 7, Choice 2 
      tmp_choice.Add_Action(new Log(Log_Name(2), R, COMMIT));
      tmp_choice.Add_Action(new Msg(Message_Name(10), 0, SND, MLT, NONE));   
      tmp_choice.Set_Next_State(2);
      tmp_choice = tmp_state.Choice(3);                   // State 7, Choice 3 
      tmp_choice.Add_Action(new Log(Log_Name(3), R, COMMIT));
      tmp_choice.Set_Next_State(4);
      tmp_choice = tmp_state.Choice(4);                   // State 7, Choice 4 
      tmp_choice.Add_Action(new Log(Log_Name(4), R, ABORT));
      tmp_choice.Set_Next_State(5);

      tmp_state = State(PART, 8);                                   // State 8
      tmp_state.Add_x_Choices(1);
      tmp_choice = tmp_state.Choice(1);                   // State 7, Choice 1 
      tmp_choice.Set_Next_State(0);
  }                                                     // void Add_Actions(0)

  protected void Display_Stats(int n) {
      // Method to output the expected statistics for protocol based on 
      // n participants
      super.Display_Statistics(n);

      Info("\tTotal messages (no failures)  \t: 6n = "+(6*n));

      // Removed - smr
      // Info("\tTotal messages (inc. failures)\t: n(6n+7)/2 ? = "
      //                                                    +(n*((6*n)+7)/2));
      Info("\tMessage rounds                \t: 6");
  }                                                   // void Display_Stats(1)

}                                                           // class Table_3PC

/*--------------------------- End of Table_3PC.java -------------------------*/

